@@ -1,7 +1,8 @@
 {-|
   Copyright   :  (C) 2015-2016, University of Twente,
                      2017-2018, Google Inc.,
-                     2021-2022, QBayLogic B.V.
+                     2021-2023, QBayLogic B.V.,
+                     2022     , Google Inc.
   License     :  BSD2 (see the file LICENSE)
   Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -48,18 +49,19 @@ import           Clash.Annotations.BitRepresentation.ClashLib
   (bitsToBits)
 import           Clash.Annotations.BitRepresentation.Util
   (BitOrigin(Lit, Field), bitOrigins, bitRanges)
-import           Clash.Core.Var                       (Attr'(..))
+import           Clash.Annotations.SynthesisAttributes (Attr(..))
 import           Clash.Debug                          (traceIf)
 import           Clash.Backend
 import           Clash.Backend.Verilog
   (bits, bit_char, encodingNote, exprLit, include, noEmptyInit, uselibs)
+import           Clash.Backend.Verilog.Time           (periodToString)
 import           Clash.Driver.Types                   (ClashOpts(..))
 import           Clash.Explicit.BlockRam.Internal     (unpackNats)
 import           Clash.Netlist.BlackBox.Types         (HdlSyn (..))
 import           Clash.Netlist.BlackBox.Util
   (extractLiterals, renderBlackBox, renderFilePath)
 import qualified Clash.Netlist.Id                     as Id
-import           Clash.Netlist.Types                  hiding (intWidth)
+import           Clash.Netlist.Types                  hiding (intWidth, usages, _usages)
 import           Clash.Netlist.Util
 import           Clash.Signal.Internal                (ActiveEdge (..))
 import           Clash.Util
@@ -93,6 +95,7 @@ data SystemVerilogState =
     , _aggressiveXOptBB_ :: AggressiveXOptBB
     , _renderEnums_ :: RenderEnums
     , _domainConfigurations_ :: DomainMap
+    , _usages :: UsageMap
     }
 
 makeLenses ''SystemVerilogState
@@ -100,6 +103,9 @@ makeLenses ''SystemVerilogState
 instance HasIdentifierSet SystemVerilogState where
   identifierSet = idSeen
 
+instance HasUsageMap SystemVerilogState where
+  usageMap = usages
+
 instance Backend SystemVerilogState where
   initBackend opts = SystemVerilogState
     { _tyCache=HashSet.empty
@@ -122,6 +128,7 @@ instance Backend SystemVerilogState where
     , _aggressiveXOptBB_=coerce (opt_aggressiveXOptBB opts)
     , _renderEnums_=coerce (opt_renderEnums opts)
     , _domainConfigurations_=emptyDomainMap
+    , _usages=mempty
     }
   hdlKind         = const SystemVerilog
   primDirs        = const $ do root <- primsRoot
@@ -197,33 +204,40 @@ type SystemVerilogM a = Ap (State SystemVerilogState) a
 
 -- | Generate SystemVerilog for a Netlist component
 genSystemVerilog
-  :: ModName
+  :: ClashOpts
+  -> ModName
   -> SrcSpan
   -> IdentifierSet
+  -> UsageMap
   -> Component
   -> SystemVerilogM ((String, Doc), [(String, Doc)])
-genSystemVerilog _ sp seen c = do
+genSystemVerilog opts _ sp seen us c = do
     -- Don't have type names conflict with module names or with previously
     -- generated type names.
     --
     -- TODO: Collect all type names up front, to prevent relatively costly union.
     -- TODO: Investigate whether type names / signal names collide in the first place
-    Ap $ idSeen %= Id.union seen
+    Ap $ do
+      idSeen %= Id.union seen
+      usages .= us
+      setSrcSpan sp
 
-    Ap $ setSrcSpan sp
     v    <- verilog
     incs <- Ap $ use includes
     return ((TextS.unpack (Id.toText cName), v), incs)
   where
     cName   = componentName c
     verilog = commentHeader <> line <>
+              nettype <> line <>
               timescale <> line <>
               module_ c
     commentHeader
          = "/* AUTOMATICALLY GENERATED SYSTEMVERILOG-2005 SOURCE CODE."
       <> line <> "** GENERATED BY CLASH " <> string (Text.pack clashVer) <> ". DO NOT MODIFY."
       <> line <> "*/"
-    timescale = "`timescale 100fs/100fs"
+    nettype = "`default_nettype none"
+    timescale = "`timescale 100fs/" <> string (Text.pack precision)
+    precision = periodToString (opt_timescalePrecision opts)
 
 -- | Generate a SystemVerilog package containing type definitions for the given HWTypes
 mkTyPackage_ :: TextS.Text -> [HWType] -> SystemVerilogM [(String,Doc)]
@@ -316,6 +330,7 @@ normaliseType ty@(Index _) = return (Unsigned (typeSize ty))
 normaliseType ty@(Sum _ _) = return (BitVector (typeSize ty))
 normaliseType ty@(CustomSum _ _ _ _) = return (BitVector (typeSize ty))
 normaliseType (Clock _) = return Bit
+normaliseType (ClockN _) = return Bit
 normaliseType (Reset _) = return Bit
 normaliseType (Enable _) = return Bool
 normaliseType (BiDirectional dir ty) = BiDirectional dir <$> normaliseType ty
@@ -380,6 +395,7 @@ splitVecTy = fmap splitElemTy . go
       Product {} -> (ns, verilogType t)
       Vector {}  -> error $ $(curLoc) ++ "impossible"
       Clock {}   -> (ns, verilogType t)
+      ClockN {}  -> (ns, verilogType t)
       Reset {}   -> (ns, "logic")
       Enable {}  -> (ns, "logic")
       Bool       -> (ns, "logic")
@@ -522,13 +538,21 @@ module_ c =
   modEnding  = "endmodule"
 
   inPorts  = sequence [ sigPort (Nothing,isBiSignalIn ty) (i,ty) Nothing | (i,ty)  <- inputs c  ]
-  outPorts = sequence [ sigPort (Just wr,False) p iEM | (wr, p, iEM) <- outputs c ]
+  outPorts = sequence [ sigPort (Just u,False) p iEM | (u, p, iEM) <- outputs c ]
+
+  -- NOTE [net types and data types]
+  --
+  -- SystemVerilog makes a distinction between the type of a net and the data
+  -- type of a signal. For output ports / inout ports this is fine, as there
+  -- is only one possible type. For input ports when using `default_nettype none
+  -- we have to specify the net type as wire explicitly (or vendor tools will
+  -- claim the net from the port declaration is implicitly defined).
 
   wr2ty (Nothing,isBidirectional)
     | isBidirectional
-    = "inout"
+    = "inout" -- no net type here, it gets added by verilogType.
     | otherwise
-    = "input"
+    = "input wire" -- See NOTE [net types and data types]
   wr2ty (Just _,_)
     = "output"
 
@@ -574,6 +598,7 @@ verilogType t_ = do
     RTree {}      -> pvrType
     Signed n      -> logicOrWire <+> "signed" <+> brackets (int (n-1) <> colon <> int 0)
     Clock _       -> "logic"
+    ClockN _      -> "logic"
     Reset _       -> "logic"
     Enable _      -> "logic"
     Bit           -> "logic"
@@ -623,6 +648,7 @@ tyName t@(Product nm _ _)      = do
 
 tyName t@(SP _ _) = "logic_vector_" <> int (typeSize t)
 tyName (Clock _)  = "logic"
+tyName (ClockN _) = "logic"
 tyName (Reset _)  = "logic"
 tyName (Enable _) = "logic"
 tyName t =  error $ $(curLoc) ++ "tyName: " ++ show t
@@ -665,35 +691,34 @@ decls ds = do
       _  -> punctuate' semi (A.pure dsDoc)
 
 decl :: Declaration -> SystemVerilogM (Maybe Doc)
-decl (NetDecl' noteM _ id_ tyE iEM) =
+decl (NetDecl' noteM id_ tyE iEM) =
   Just A.<$> maybe id addNote noteM (addAttrs attrs (typ tyE))
   where
-    typ (Left  ty) = stringS ty <+> pretty id_ <> iE
-    typ (Right ty) = sigDecl (pretty id_) ty <> iE
+    typ ty = sigDecl (pretty id_) ty <> iE
     addNote n = mappend ("//" <+> stringS n <> line)
-    attrs = fromMaybe [] (hwTypeAttrs A.<$> either (const Nothing) Just tyE)
+    attrs = fromMaybe [] (hwTypeAttrs A.<$> Just tyE)
     iE = maybe emptyDoc (noEmptyInit . expr_ False) iEM
 
 decl _ = return Nothing
 
 -- | Convert single attribute to systemverilog syntax
-renderAttr :: Attr' -> Text.Text
-renderAttr (StringAttr'  key value) = Text.pack $ concat [key, " = ", show value]
-renderAttr (IntegerAttr' key value) = Text.pack $ concat [key, " = ", show value]
-renderAttr (BoolAttr'    key True ) = Text.pack $ concat [key, " = ", "1"]
-renderAttr (BoolAttr'    key False) = Text.pack $ concat [key, " = ", "0"]
-renderAttr (Attr'        key      ) = Text.pack $ key
+renderAttr :: Attr TextS.Text -> TextS.Text
+renderAttr (StringAttr  key value) = TextS.concat [key, " = ", TextS.pack (show value)]
+renderAttr (IntegerAttr key value) = TextS.concat [key, " = ", TextS.pack (show value)]
+renderAttr (BoolAttr    key True ) = TextS.concat [key, " = ", "1"]
+renderAttr (BoolAttr    key False) = TextS.concat [key, " = ", "0"]
+renderAttr (Attr        key      ) = key
 
 -- | Add attribute notation to given declaration
 addAttrs
-  :: [Attr']
+  :: [Attr TextS.Text]
   -> SystemVerilogM Doc
   -> SystemVerilogM Doc
 addAttrs []     t = t
 addAttrs attrs' t =
   "(*" <+> attrs'' <+> "*)" <+> t
     where
-      attrs'' = string $ Text.intercalate ", " (map renderAttr attrs')
+      attrs'' = stringS $ TextS.intercalate ", " (map renderAttr attrs')
 
 insts :: [Declaration] -> SystemVerilogM Doc
 insts [] = emptyDoc
@@ -777,7 +802,9 @@ inst_' id_ scrut scrutTy es = fmap Just $
 inst_ :: Declaration -> SystemVerilogM (Maybe Doc)
 inst_ (TickDecl {}) = return Nothing
 
-inst_ (Assignment id_ e) = fmap Just $
+inst_ (CompDecl {}) = return Nothing
+
+inst_ (Assignment id_ Cont e) = fmap Just $
   "assign" <+> pretty id_ <+> equals <+> align (expr_ False e <> semi)
 
 inst_ (CondAssignment id_ ty scrut _ [(Just (BoolLit b), l),(_,r)]) = fmap Just $ do
@@ -868,6 +895,9 @@ inst_ (NetDecl' {}) = return Nothing
 inst_ (ConditionalDecl cond ds) = Just <$>
   "`ifdef" <+> pretty cond <> line <> indent 2 (insts ds) <> line <> "`endif"
 
+inst_ d =
+  error ("inst_: " ++ show d)
+
 -- | Render a data constructor application for data constructors having a
 -- custom bit representation.
 customReprDataCon
@@ -952,8 +982,9 @@ seq_ (Branch scrut scrutTy es) =
           "end") <:> conds es'
 
 seq_ (SeqDecl sd) = case sd of
-  Assignment id_ e ->
-    pretty id_ <+> equals <+> expr_ False e <> semi
+  Assignment id_ (Proc b) e ->
+    let sym = case b of { Blocking -> equals; NonBlocking -> "<=" }
+     in pretty id_ <+> sym <+> expr_ False e <> semi
 
   BlackBoxD {} ->
     fromMaybe <$> emptyDoc <*> inst_ sd
@@ -1099,7 +1130,7 @@ expr_ _ (Identifier id_ (Just (Indexed ((Unsigned w),_,_))))  = do
 expr_ _ (Identifier _ (Just (Indexed ((BitVector _),_,0)))) = do
   iw <- Ap $ use intWidth
   traceIf True ($(curLoc) ++ "WARNING: synthesizing bitvector mask to dontcare") $
-    verilogTypeErrValue (Signed iw)
+    verilogTypeErrValue (Unsigned iw)
 
 -- See [Note] bitvector projection
 expr_ _ (Identifier id_ (Just (Indexed ((BitVector w),_,1)))) = do
@@ -1116,10 +1147,10 @@ expr_ b (DataCon _ (DC (Void {}, -1)) [e]) =  expr_ b e
 
 expr_ _ (DataCon ty@(Vector 0 _) _ _) = verilogTypeErrValue ty
 
-expr_ _ (DataCon (Vector 1 elTy) _ [e]) = "'" <> braces (toSLV elTy e)
+expr_ _ (DataCon (Vector 1 elTy) _ [e]) = "'" <> braces (int 0 <> colon <+> toSLV elTy e)
 
 expr_ _ e@(DataCon ty@(Vector _ elTy) _ [e1,e2]) = case vectorChain e of
-  Just es -> "'" <> listBraces (mapM (toSLV elTy) es)
+  Just es -> "'" <> listBraces (zipWithM (\i e3 -> int i <> colon <+> toSLV elTy e3) [0..] es)
   Nothing -> verilogTypeMark ty <> "_cons" <> parens (expr_ False e1 <> comma <+> expr_ False e2)
 
 expr_ _ (DataCon (MemBlob n m) _ [n0, m0, _, runs, _, ends])
@@ -1190,9 +1221,9 @@ expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
 expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
   | pNm == "Clash.Sized.Internal.BitVector.fromInteger##"
   , [Literal _ m, Literal _ i] <- extractLiterals bbCtx
-  = let NumLit m' = m
-        NumLit i' = i
-    in exprLitSV (Just (Bit,1)) (BitLit $ toBit m' i')
+  , NumLit m' <- m
+  , NumLit i' <- i
+  = exprLitSV (Just (Bit,1)) (BitLit $ toBit m' i')
 
 expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
   | pNm == "Clash.Sized.Internal.Index.fromInteger#"
