@@ -1,7 +1,8 @@
 {-|
   Copyright   :  (C) 2015-2016, University of Twente,
                      2017-2018, Google Inc.,
-                     2021-2022, QBayLogic B.V.
+                     2021-2023, QBayLogic B.V.
+                     2022     , Google Inc.
   License     :  BSD2 (see the file LICENSE)
   Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -20,7 +21,9 @@
 module Clash.Backend.VHDL (VHDLState) where
 
 import           Control.Arrow                        (second)
+#if !MIN_VERSION_base(4,18,0)
 import           Control.Applicative                  (liftA2)
+#endif
 import           Control.Lens                         hiding (Indexed, Empty)
 import           Control.Monad                        (forM,join,zipWithM)
 import           Control.Monad.State                  (State, StateT)
@@ -35,7 +38,7 @@ import qualified Data.HashMap.Strict                  as HashMapS
 import           Data.HashSet                         (HashSet)
 import qualified Data.HashSet                         as HashSet
 import           Data.List
-  (mapAccumL, nub, nubBy, intersperse, group, sort)
+  (mapAccumL, nub, nubBy, partition, intersperse, group, sort)
 import           Data.List.Extra                      ((<:>), equalLength, zipEqual)
 import           Data.Maybe                           (catMaybes,mapMaybe)
 import           Data.Monoid                          (Ap(Ap))
@@ -62,8 +65,8 @@ import           Clash.Annotations.BitRepresentation.ClashLib
   (bitsToBits)
 import           Clash.Annotations.BitRepresentation.Util
   (BitOrigin(Lit, Field), bitOrigins, bitRanges)
+import           Clash.Annotations.SynthesisAttributes (Attr(..))
 import           Clash.Backend
-import           Clash.Core.Var                       (Attr'(..),attrName)
 import           Clash.Debug                          (traceIf)
 import           Clash.Driver.Types                   (ClashOpts(..))
 import           Clash.Explicit.BlockRam.Internal     (unpackNats)
@@ -71,7 +74,7 @@ import           Clash.Netlist.BlackBox.Types         (HdlSyn (..))
 import           Clash.Netlist.BlackBox.Util
   (extractLiterals, renderBlackBox, renderFilePath)
 import qualified Clash.Netlist.Id                     as Id
-import           Clash.Netlist.Types                  hiding (intWidth)
+import           Clash.Netlist.Types                  hiding (intWidth, usages, _usages)
 import           Clash.Netlist.Util
 import           Clash.Util
   (SrcSpan, noSrcSpan, clogBase, curLoc, makeCached, indexNote)
@@ -115,6 +118,7 @@ data VHDLState =
   , _aggressiveXOptBB_ :: AggressiveXOptBB
   , _renderEnums_ :: RenderEnums
   , _domainConfigurations_ :: DomainMap
+  , _usages :: UsageMap
   }
 
 makeLenses ''VHDLState
@@ -122,6 +126,9 @@ makeLenses ''VHDLState
 instance HasIdentifierSet VHDLState where
   identifierSet = idSeen
 
+instance HasUsageMap VHDLState where
+  usageMap = usages
+
 instance Backend VHDLState where
   initBackend opts = VHDLState
     { _tyCache=mempty
@@ -144,6 +151,7 @@ instance Backend VHDLState where
     , _aggressiveXOptBB_=coerce (opt_aggressiveXOptBB opts)
     , _renderEnums_=coerce (opt_renderEnums opts)
     , _domainConfigurations_=emptyDomainMap
+    , _usages=mempty
     }
   hdlKind         = const VHDL
   primDirs        = const $ do root <- primsRoot
@@ -169,6 +177,7 @@ instance Backend VHDLState where
       if enums then pure UserType else pure SynonymType
 
     Clock {} -> pure SynonymType
+    ClockN {} -> pure SynonymType
     Reset {} -> pure SynonymType
     Enable {} -> pure SynonymType
     Index {} -> pure SynonymType
@@ -242,20 +251,28 @@ instance Backend VHDLState where
   blockDecl nm ds = do
     decs <- decls ds
     let attrs = [ (id_, attr)
-                | NetDecl' _ _ id_ (Right hwtype) _ <- ds
+                | NetDecl' _ id_ hwtype _ <- ds
                 , attr <- hwTypeAttrs hwtype]
     if isEmpty decs
-       then insts ds
+       then insts ids
        else nest 2
               (pretty nm <+> colon <+> "block" <> line <>
                pure decs <>
-               if null attrs
+               (if null attrs
                 then emptyDoc
-                else line <> line <> renderAttrs (TextS.pack "signal") attrs) <> line <>
-            nest 2
-              ("begin" <> line <>
-                insts ds) <> line <>
+                else line <> line <> renderAttrs (TextS.pack "signal") attrs) <>
+               if null cds
+                then emptyDoc
+                else line <> line <> insts cds) <> line <> "begin" <>
+            nest 2 (line <> insts ids)
+            <> line <>
             "end block" <> semi
+   where
+     (cds, ids) = partition isCompDecl ds
+
+     isCompDecl (CompDecl {}) = True
+     isCompDecl _             = False
+
   addIncludes inc = includes %= (inc++)
   addLibraries libs = libraries %= (libs ++)
   addImports imps = packages %= (imps ++)
@@ -282,7 +299,7 @@ isBV _ _ = False
 
 -- | Generate unique (partial) names for product fields. Example:
 --
--- >>> productFieldNames [Unsigned 6, Unsigned 6, Bit, Bool]
+-- > productFieldNames Nothing [Unsigned 6, Unsigned 6, Bit, Bool]
 -- ["unsigned6_0", "unsigned6_1", "bit", "boolean"]
 productFieldNames
   :: HasCallStack
@@ -296,7 +313,9 @@ productFieldNames labels0 fields = do
   hFields <- zipWithM hName labels1 fields
 
   let grouped = group $ sort $ hFields
-      counted = HashMapS.fromList (map (\(g:gs) -> (g, succ (length gs))) grouped)
+      countGroup [] = error "productFIeldNames.countGroup: group of zero elements"
+      countGroup (g:gs) = (g, succ (length gs))
+      counted = HashMapS.fromList (map countGroup grouped)
       names   = snd $ mapAccumL (name' counted) HashMapS.empty hFields
 
   return names
@@ -373,20 +392,24 @@ enumVariantName _ _ =
 
 -- | Generate VHDL for a Netlist component
 genVHDL
-  :: ModName
+  :: ClashOpts
+  -> ModName
   -> SrcSpan
   -> IdentifierSet
+  -> UsageMap
   -> Component
   -> VHDLM ((String, Doc), [(String, Doc)])
-genVHDL nm sp seen c = do
+genVHDL _ nm sp seen us c = do
     -- Don't have type names conflict with module names or with previously
     -- generated type names.
     --
     -- TODO: Collect all type names up front, to prevent relatively costly union.
     -- TODO: Investigate whether type names / signal names collide in the first place
-    Ap $ idSeen %= Id.union seen
+    Ap $ do
+      idSeen %= Id.union seen
+      usages .= us
+      setSrcSpan sp
 
-    Ap $ setSrcSpan sp
     v <- vhdl
     i <- Ap $ use includes
     Ap $ libraries .= []
@@ -579,6 +602,7 @@ tyDec hwty = do
 
     -- Type aliases:
     Clock _           -> typAliasDec hwty
+    ClockN _          -> typAliasDec hwty
     Reset _           -> typAliasDec hwty
     Enable _          -> typAliasDec hwty
     Index _           -> typAliasDec hwty
@@ -721,8 +745,9 @@ funDec _ _ t@(Product _ labels elTys) = Just
                             parens ("p." <> tyName t <> selectProductField labels elTys i))
 
     argLengths = map typeSize elTys
-    starts     = 0 : snd (mapAccumL ((join (,) .) . (+)) 0 argLengths)
-    ends       = map (subtract 1) (tail starts)
+    starts1    = snd (mapAccumL ((join (,) .) . (+)) 0 argLengths)
+    starts     = 0 : starts1
+    ends       = map (subtract 1) starts1
 
     elTyFromSLV = forM (zip starts ends)
                        (\(s,e) -> "fromSLV" <>
@@ -940,7 +965,7 @@ entity c = do
     rattrs      = renderAttrs (TextS.pack "signal") attrs
     attrs       = inputAttrs ++ outputAttrs
     inputAttrs  = [(id_, attr) | (id_, hwtype) <- inputs c, attr <- hwTypeAttrs hwtype]
-    outputAttrs = [(id_, attr) | (_wireOrReg, (id_, hwtype), _) <- outputs c, attr <- hwTypeAttrs hwtype]
+    outputAttrs = [(id_, attr) | (_, (id_, hwtype), _) <- outputs c, attr <- hwTypeAttrs hwtype]
 
 
 architecture :: Component -> VHDLM Doc
@@ -961,46 +986,52 @@ architecture c = do {
   }
  where
    netdecls    = filter isNetDecl (declarations c)
-   declAttrs   = [(id_, attr) | NetDecl' _ _ id_ (Right hwtype) _ <- netdecls, attr <- hwTypeAttrs hwtype]
+   declAttrs   = [(id_, attr) | NetDecl' _ id_ hwtype _ <- netdecls, attr <- hwTypeAttrs hwtype]
    inputAttrs  = [(id_, attr) | (id_, hwtype) <- inputs c, attr <- hwTypeAttrs hwtype]
-   outputAttrs = [(id_, attr) | (_wireOrReg, (id_, hwtype), _) <- outputs c, attr <- hwTypeAttrs hwtype]
+   outputAttrs = [(id_, attr) | (_, (id_, hwtype), _) <- outputs c, attr <- hwTypeAttrs hwtype]
 
    isNetDecl :: Declaration -> Bool
-   isNetDecl (NetDecl' _ _ _ (Right _) _) = True
-   isNetDecl _                            = False
+   isNetDecl NetDecl'{} = True
+   isNetDecl _          = False
 
-attrType
-  :: t ~ HashMap T.Text T.Text
-  => t
-  -> Attr'
-  -> t
+attrType ::
+  HashMap TextS.Text TextS.Text ->
+  Attr TextS.Text ->
+  HashMap TextS.Text TextS.Text
 attrType types attr =
   case HashMap.lookup name' types of
     Nothing    -> HashMap.insert name' type' types
     Just type'' | type'' == type' -> types
                 | otherwise -> error $
-                      $(curLoc) ++ unwords [ T.unpack name', "already assigned"
-                                           , T.unpack type'', "while we tried to"
-                                           , "add", T.unpack type' ]
+                      $(curLoc) ++ unwords [ TextS.unpack name', "already assigned"
+                                           , TextS.unpack type'', "while we tried to"
+                                           , "add", TextS.unpack type' ]
  where
-  name' = T.pack $ attrName attr
-  type' = T.pack $ case attr of
-            BoolAttr' _ _    -> "boolean"
-            IntegerAttr' _ _ -> "integer"
-            StringAttr' _ _  -> "string"
-            Attr' _          -> "bool"
+  name' = attrName attr
+  type' = case attr of
+            BoolAttr _ _    -> "boolean"
+            IntegerAttr _ _ -> "integer"
+            StringAttr _ _  -> "string"
+            Attr _          -> "boolean"
+
+attrName :: Attr a -> a
+attrName = \case
+  BoolAttr a _ -> a
+  IntegerAttr a _ -> a
+  StringAttr a _ -> a
+  Attr a -> a
 
 -- | Create 'attrname -> type' mapping for given attributes. Will err if multiple
 -- types are assigned to the same name.
-attrTypes :: [Attr'] -> HashMap T.Text T.Text
+attrTypes :: [Attr TextS.Text] -> HashMap TextS.Text TextS.Text
 attrTypes = foldl attrType HashMap.empty
 
 -- | Create a 'attrname -> (type, [(signalname, value)]). Will err if multiple
 -- types are assigned to the same name.
 attrMap
   :: forall t
-   . t ~ HashMap T.Text (T.Text, [(TextS.Text, T.Text)])
-  => [(Identifier, Attr')]
+   . t ~ HashMap TextS.Text (TextS.Text, [(TextS.Text, TextS.Text)])
+  => [(Identifier, Attr TextS.Text)]
   -> t
 attrMap attrs0 = foldl go empty' attrs1
  where
@@ -1010,55 +1041,56 @@ attrMap attrs0 = foldl go empty' attrs1
            [(k, (types HashMap.! k, [])) | k <- HashMap.keys types]
   types = attrTypes (map snd attrs1)
 
-  go :: t -> (TextS.Text, Attr') -> t
-  go map' attr = HashMap.adjust
-                   (go' attr)
-                   (T.pack $ attrName $ snd attr)
-                   map'
+  go :: t -> (TextS.Text, Attr TextS.Text) -> t
+  go map' attr = HashMap.adjust (go' attr) (attrName $ snd attr) map'
 
   go'
-    :: (TextS.Text, Attr')
-    -> (T.Text, [(TextS.Text, T.Text)])
-    -> (T.Text, [(TextS.Text, T.Text)])
+    :: (TextS.Text, Attr TextS.Text)
+    -> (TextS.Text, [(TextS.Text, TextS.Text)])
+    -> (TextS.Text, [(TextS.Text, TextS.Text)])
   go' (signalName, attr) (typ, elems) =
     (typ, (signalName, renderAttr attr) : elems)
 
 renderAttrs
   :: TextS.Text
-  -> [(Identifier, Attr')]
+  -> [(Identifier, Attr TextS.Text)]
   -> VHDLM Doc
 renderAttrs what (attrMap -> attrs) =
   vcat $ sequence $ intersperse " " $ map renderAttrGroup (HashMap.toList attrs)
  where
   renderAttrGroup
-    :: (T.Text, (T.Text, [(TextS.Text, T.Text)]))
+    :: (TextS.Text, (TextS.Text, [(TextS.Text, TextS.Text)]))
     -> VHDLM Doc
   renderAttrGroup (attrname, (typ, elems)) =
-    ("attribute" <+> string attrname <+> colon <+> string typ <> semi)
+    ("attribute" <+> stringS attrname <+> colon <+> stringS typ <> semi)
     <> line <>
     (vcat $ sequence $ map (renderAttrDecl attrname) elems)
 
   renderAttrDecl
-    :: T.Text
-    -> (TextS.Text, T.Text)
+    :: TextS.Text
+    -> (TextS.Text, TextS.Text)
     -> VHDLM Doc
   renderAttrDecl attrname (signalName, value) =
         "attribute"
-    <+> string attrname
+    <+> stringS attrname
     <+> "of"
     <+> stringS signalName -- or component name
     <+> colon
     <+> stringS what <+> "is" -- "signal is" or "component is"
-    <+> string value
+    <+> stringS value
     <> semi
 
 -- | Convert single attribute to VHDL syntax
-renderAttr :: Attr' -> T.Text
-renderAttr (StringAttr'  _key value) = T.replace "\\\"" "\"\"" $ T.pack $ show value
-renderAttr (IntegerAttr' _key value) = T.pack $ show value
-renderAttr (BoolAttr'    _key True ) = T.pack $ "true"
-renderAttr (BoolAttr'    _key False) = T.pack $ "false"
-renderAttr (Attr'        _key      ) = T.pack $ "true"
+renderAttr :: Attr TextS.Text -> TextS.Text
+renderAttr (StringAttr  _key value) = wrap '"' $ TextS.replace "\"" "\"\"" value
+renderAttr (IntegerAttr _key value) = TextS.pack (show value)
+renderAttr (BoolAttr    _key True ) = "true"
+renderAttr (BoolAttr    _key False) = "false"
+renderAttr (Attr        _key      ) = "true"
+
+-- | Prepend and append a character to a string
+wrap :: Char -> TextS.Text -> TextS.Text
+wrap c = cons c . (`snoc` c)
 
 sigDecl :: VHDLM Doc -> HWType -> VHDLM Doc
 sigDecl d t = d <+> colon <+> sizedQualTyName t
@@ -1173,6 +1205,9 @@ tyName' rec0 (filterTransparent -> t) = do
     Clock nm0 ->
       let nm1 = "clk_" `TextS.append` nm0 in
       Ap $ makeCached (t, False) nameCache (userTyName "clk" nm1 t)
+    ClockN nm0 ->
+      let nm1 = "clk_n_" `TextS.append` nm0 in
+      Ap $ makeCached (t, False) nameCache (userTyName "clk" nm1 t)
     Reset nm0 ->
       let nm1 = "rst_" `TextS.append` nm0 in
       Ap $ makeCached (t, False) nameCache (userTyName "rst" nm1 t)
@@ -1225,6 +1260,7 @@ normaliseType enums@(RenderEnums e) hwty = case hwty of
 
   -- Simple types, for which a subtype (without qualifiers) will be made in VHDL:
   Clock _           -> Bit
+  ClockN _          -> Bit
   Reset _           -> Bit
   Enable _          -> Bool
   Index _           -> Unsigned (typeSize hwty)
@@ -1248,6 +1284,7 @@ filterTransparent hwty = case hwty of
   Integer           -> hwty
   Bit               -> hwty
   Clock _           -> hwty
+  ClockN _          -> hwty
   Reset _           -> hwty
   Enable _          -> hwty
   Index _           -> hwty
@@ -1275,8 +1312,8 @@ filterTransparent hwty = case hwty of
       (map (second filterTransparent) constrs)
 
   -- Transparent types:
-  Annotated _ elTy -> elTy
-  BiDirectional _ elTy -> elTy
+  Annotated _ elTy -> filterTransparent elTy
+  BiDirectional _ elTy -> filterTransparent elTy
 
   Void {} -> hwty
   KnownDomain {} -> hwty
@@ -1326,6 +1363,7 @@ sizedQualTyNameErrValue t@(Sum _ _)  = do
   else
     qualTyName t <> "'" <> parens (int 0 <+> "to" <+> int (typeSize t - 1) <+> rarrow <+> singularErrValue)
 sizedQualTyNameErrValue (Clock _)  = singularErrValue
+sizedQualTyNameErrValue (ClockN _) = singularErrValue
 sizedQualTyNameErrValue (Reset _)  = singularErrValue
 sizedQualTyNameErrValue (Enable _) = singularErrValue
 sizedQualTyNameErrValue (Void {})  =
@@ -1360,8 +1398,8 @@ decls ds = do
       _  -> vcat (pure dsDoc)
 
 decl :: Int ->  Declaration -> VHDLM (Maybe (Doc,Int))
-decl l (NetDecl' noteM _ id_ ty iEM) = Just <$> (,fromIntegral (TextS.length (Id.toText id_))) <$>
-  maybe id addNote noteM ("signal" <+> fill l (pretty id_) <+> colon <+> either pretty sizedQualTyName ty <> iE <> semi)
+decl l (NetDecl' noteM id_ ty iEM) = Just <$> (,fromIntegral (TextS.length (Id.toText id_))) <$>
+  maybe id addNote noteM ("signal" <+> fill l (pretty id_) <+> colon <+> sizedQualTyName ty <> iE <> semi)
   where
     addNote n = mappend ("--" <+> pretty n <> line)
     iE = maybe emptyDoc (noEmptyInit . expr_ False) iEM
@@ -1384,7 +1422,9 @@ decl _ (InstDecl Comp _ attrs nm _ gens (NamedPortMap pms)) = fmap (Just . (,0))
     portDir In  = "in"
     portDir Out = "out"
 
-    attrs' = if null attrs then emptyDoc else renderAttrs (TextS.pack "component") [(nm, a) | a <- attrs]
+    attrs'
+      | null attrs = emptyDoc
+      | otherwise = renderAttrs (TextS.pack "component") [(nm, a) | a <- attrs]
 
 decl _ _ = return Nothing
 
@@ -1479,9 +1519,16 @@ inst_' id_ scrut scrutTy es = fmap Just $
 
 -- | Turn a Netlist Declaration to a VHDL concurrent block
 inst_ :: Declaration -> VHDLM (Maybe Doc)
-inst_ (Assignment id_ e) = fmap Just $
+inst_ (Assignment id_ Cont e) = fmap Just $
   pretty id_ <+> larrow <+> align (expr_ False e) <> semi
 
+inst_ (CompDecl nm ps0) =
+  fmap Just $ "component" <+> pretty nm <+>
+    ("port" <> line <> indent 2 (tupledSemi ps <> semi))
+    <> line <> "end component" <> semi
+  where ps = traverse (\(t,pd,ty) -> pretty t <+> ":" <+> ppd pd <+> sizedQualTyName ty) ps0
+        ppd = \case { In -> "in"; Out -> "out"}
+
 inst_ (CondAssignment id_ _ scrut _ [(Just (BoolLit b), l),(_,r)]) = fmap Just $
   pretty id_ <+> larrow
            <+> align (vsep (sequence [expr_ False t <+> "when" <+>
@@ -1706,16 +1753,16 @@ expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
 expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
   | pNm == "Clash.Sized.Internal.BitVector.fromInteger#"
   , [Literal _ (NumLit n), Literal _ m, Literal _ i] <- extractLiterals bbCtx
-  = let NumLit m' = m
-        NumLit i' = i
-    in exprLit (Just (BitVector (fromInteger n),fromInteger n)) (BitVecLit m' i')
+  , NumLit m' <- m
+  , NumLit i' <- i
+  = exprLit (Just (BitVector (fromInteger n),fromInteger n)) (BitVecLit m' i')
 
 expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
   | pNm == "Clash.Sized.Internal.BitVector.fromInteger##"
   , [Literal _ m, Literal _ i] <- extractLiterals bbCtx
-  = let NumLit m' = m
-        NumLit i' = i
-    in exprLit (Just (Bit,1)) (BitLit $ toBit m' i')
+  , NumLit m' <- m
+  , NumLit i' <- i
+  = exprLit (Just (Bit,1)) (BitLit $ toBit m' i')
 
 expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
   | pNm == "Clash.Sized.Internal.Index.fromInteger#"
@@ -1890,9 +1937,9 @@ bits :: [Bit] -> VHDLM Doc
 bits = dquotes . hcat . mapM bit_char
 
 toHex :: Int -> Integer -> String
-toHex sz i =
-  let Just d = clogBase 16 (2^sz)
-  in  printf ("%0" ++ show d ++ "X") (abs i)
+toHex sz i = case clogBase 16 (2^sz) of
+  Just d -> printf ("%0" ++ show d ++ "X") (abs i)
+  _ -> error "toHex: impossible"
 
 hex :: String -> VHDLM Doc
 hex s = char 'x' <> dquotes (pretty (T.pack s))
@@ -1905,7 +1952,7 @@ bit_char U = do
   case udf of
     Nothing -> char '-'
     Just Nothing -> char '0'
-    Just (Just i) -> "'" <> int i <> "'"
+    Just (Just i) -> int i
 bit_char Z = char 'Z'
 
 toSLV :: HasCallStack => HWType -> Expr -> VHDLM Doc
@@ -1918,6 +1965,9 @@ toSLV Bit          e = do
 toSLV (Clock {})    e = do
   nm <- Ap $ use modNm
   pretty nm <> "_types.toSLV" <> parens (expr_ False e)
+toSLV (ClockN {})   e = do
+  nm <- Ap $ use modNm
+  pretty nm <> "_types.toSLV" <> parens (expr_ False e)
 toSLV (Reset {})    e = do
   nm <- Ap $ use modNm
   pretty (TextS.toLower nm) <> "_types.toSLV" <> parens (expr_ False e)
@@ -2004,6 +2054,7 @@ punctuate' s d = vcat (punctuate s d) <> s
 
 encodingNote :: HWType -> VHDLM Doc
 encodingNote (Clock _)  = "-- clock" <> line
+encodingNote (ClockN _) = "-- clock (neg phase)" <> line
 encodingNote (Reset _)  = "-- reset" <> line
 encodingNote (Enable _) = "-- enable" <> line
 encodingNote (Annotated _ t) = encodingNote t
