@@ -1,7 +1,8 @@
 {-|
   Copyright   :  (C) 2015-2016, University of Twente,
                      2017-2018, Google Inc.,
-                     2021-2022, QBayLogic B.V.
+                     2021-2023, QBayLogic B.V.
+                     2022     , Google Inc.
   License     :  BSD2 (see the file LICENSE)
   Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -64,8 +65,9 @@ import           Clash.Annotations.BitRepresentation.Internal
   (ConstrRepr'(..), DataRepr'(..), ConstrRepr'(..))
 import           Clash.Annotations.BitRepresentation.Util
   (BitOrigin(Lit, Field), bitOrigins, bitRanges)
-import           Clash.Core.Var                       (Attr'(..))
+import           Clash.Annotations.SynthesisAttributes (Attr(..))
 import           Clash.Backend
+import           Clash.Backend.Verilog.Time           (periodToString)
 import           Clash.Debug                          (traceIf)
 import           Clash.Driver.Types                   (ClashOpts(..))
 import           Clash.Explicit.BlockRam.Internal     (unpackNats)
@@ -73,7 +75,7 @@ import           Clash.Netlist.BlackBox.Types         (HdlSyn)
 import           Clash.Netlist.BlackBox.Util
   (extractLiterals, renderBlackBox, renderFilePath)
 import qualified Clash.Netlist.Id                     as Id
-import           Clash.Netlist.Types                  hiding (intWidth)
+import           Clash.Netlist.Types as N             hiding (intWidth, usages, _usages)
 import           Clash.Netlist.Util
 import           Clash.Signal.Internal                (ActiveEdge (..))
 import           Clash.Util
@@ -101,6 +103,7 @@ data VerilogState =
     , _undefValue :: Maybe (Maybe Int)
     , _aggressiveXOptBB_ :: AggressiveXOptBB
     , _domainConfigurations_ :: DomainMap
+    , _usages :: UsageMap
     }
 
 makeLenses ''VerilogState
@@ -108,6 +111,9 @@ makeLenses ''VerilogState
 instance HasIdentifierSet VerilogState where
   identifierSet = idSeen
 
+instance HasUsageMap VerilogState where
+  usageMap = usages
+
 instance Backend VerilogState where
   initBackend opts = VerilogState
     { _genDepth=0
@@ -125,6 +131,7 @@ instance Backend VerilogState where
     , _undefValue=opt_forceUndefined opts
     , _aggressiveXOptBB_=coerce (opt_aggressiveXOptBB opts)
     , _domainConfigurations_=emptyDomainMap
+    , _usages=mempty
     }
   hdlKind         = const Verilog
   primDirs        = const $ do root <- primsRoot
@@ -136,7 +143,7 @@ instance Backend VerilogState where
   name            = const "verilog"
   extension       = const ".v"
 
-  genHDL          = const genVerilog
+  genHDL          = genVerilog
   mkTyPackage _ _ = return []
   hdlType _       = verilogType
   hdlHWTypeKind _ = pure PrimitiveType -- Everything is a bitvector!
@@ -197,20 +204,25 @@ type VerilogM a = Ap (State VerilogState) a
 
 -- | Generate Verilog for a Netlist component
 genVerilog
-  :: SrcSpan
+  :: ClashOpts
+  -> ModName
+  -> SrcSpan
   -> IdentifierSet
+  -> UsageMap
   -> Component
   -> VerilogM ((String, Doc), [(String, Doc)])
-genVerilog sp seen c = do
+genVerilog opts  _ sp seen usage c = do
     -- Don't have type names conflict with module names or with previously
     -- generated type names.
     --
     -- TODO: Collect all type names up front, to prevent relatively costly union.
     -- TODO: Investigate whether type names / signal names collide in the first place
-    Ap $ idSeen %= Id.union seen
+    Ap $ do
+      idSeen %= Id.union seen
+      usages .= usage
+      setSrcSpan sp
 
-    Ap (setSrcSpan sp)
-    v    <- commentHeader <> line <> timescale <> line <> module_ c
+    v    <- commentHeader <> line <> nettype <> line <> timescale <> line <> module_ c
     incs <- Ap $ use includes
     return ((TextS.unpack (Id.toText cName), v), incs)
   where
@@ -219,25 +231,76 @@ genVerilog sp seen c = do
          = "/* AUTOMATICALLY GENERATED VERILOG-2001 SOURCE CODE."
       <> line <> "** GENERATED BY CLASH " <> string (Text.pack clashVer) <> ". DO NOT MODIFY."
       <> line <> "*/"
-    timescale = "`timescale 100fs/100fs"
+    nettype = "`default_nettype none"
+    timescale = "`timescale 100fs/" <> string (Text.pack precision)
+    precision = periodToString (opt_timescalePrecision opts)
 
 sigPort
-  :: Maybe WireOrReg
+  :: VerilogM Doc
+  -> Maybe N.Usage
   -> Identifier
   -> HWType
   -> Maybe Expr
   -> VerilogM Doc
-sigPort wor (Id.toText -> pName) hwType iEM =
+sigPort def mu (Id.toText -> pName) hwType iEM = do
     addAttrs (hwTypeAttrs hwType)
       (portType <+> verilogType hwType <+> stringS pName <> iE <> encodingNote hwType)
   where
-    portType = case wor of
-                 Nothing   -> if isBiSignalIn hwType then "inout" else "input"
-                 Just Wire -> "output" <+> "wire"
-                 Just Reg  -> "output" <+> "reg"
+    portType =
+      -- See NOTE [net and variable ports]
+      case mu of
+        Just Cont ->
+          "output wire"
+
+        Just Proc{} ->
+          "output reg"
+
+        Nothing ->
+          if isBiSignalIn hwType then "inout wire" else def <+> "wire"
 
     iE = maybe emptyDoc (noEmptyInit . expr_ False) iEM
 
+{-
+NOTE [net and variable ports]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+In Verilog, ports are typically seen written with the default implicit net type
+of wire, i.e.
+
+  input foo,
+  output bar,
+
+which is really a shorthand for
+
+  input wire foo,
+  output wire bar,
+
+When we use `default_nettype none however, this is no longer allowed as all
+nets must be explicitly given their type. When generating code we must include
+the additional word wire (the net type).
+
+What is the benefit of this? When we have a default net type, any variable
+generated which is not declared is given an implicit declaration of this
+default type. This can obscure clash errors, i.e. if Clash generated
+
+  wire foo;
+  fo <= ...;
+
+Then simulators would act as though this was written:
+
+  wire foo;
+  wire fo;
+  fo <= ...;
+
+Which should never be the desired behaviour for code generated by Clash.
+
+One final point, Verilog allows output to also be a variable type, i.e.
+
+  output reg foo;
+
+If a port is an input or an inout then it can only be a net and not a variable
+according to the standard, so input reg and inout reg are impossible.
+-}
+
 module_ :: Component -> VerilogM Doc
 module_ c =
   modVerilog <* Ap (imports .= HashSet.empty >> libraries .= HashSet.empty)
@@ -256,8 +319,11 @@ module_ c =
     modBody    = indent 2 (decls (declarations c)) <> line <> line <> indent 2 (insts (declarations c))
     modEnding  = "endmodule"
 
-    inPorts  = sequence [ sigPort Nothing id_ hwType Nothing | (id_, hwType) <- inputs c  ]
-    outPorts = sequence [ sigPort (Just wireOrReg) id_ hwType iEM | (wireOrReg, (id_, hwType), iEM) <- outputs c ]
+    inPorts  = sequence [ sigPort "input" Nothing id_ hwType Nothing | (id_, hwType) <- inputs c  ]
+    outPorts = do
+      us <- use usages
+      let useOf i u = lookupUsage i us <> Just u
+      sequence [ sigPort "output" (useOf id_ u) id_ hwType iEM | (u,(id_, hwType), iEM) <- outputs c ]
 
     -- slightly more readable than 'tupled', makes the output Haskell-y-er
     commafy v = (comma <> space) <> pure v
@@ -291,21 +357,23 @@ uselibs xs = line <>
   indent 2 (string "`uselib" <+> (hsep (mapM (\l -> ("lib=" <> string l)) xs)))
   <> line <> line
 
-wireRegFileDoc :: WireOrReg -> (Either a HWType) -> VerilogM Doc
-wireRegFileDoc _    (Right FileType) = "integer"
-wireRegFileDoc Wire _                = "wire"
-wireRegFileDoc Reg  _                = "reg"
+usageFileDoc :: Maybe N.Usage -> HWType -> VerilogM Doc
+usageFileDoc _             FileType  = "integer"
+usageFileDoc (Just Proc{}) _         = "reg"
+usageFileDoc _             _         = "wire"
 
 verilogType :: HWType -> VerilogM Doc
 verilogType t = case t of
   Signed n -> "signed" <+> brackets (int (n-1) <> colon <> int 0)
   Clock {} -> emptyDoc
+  ClockN {} -> emptyDoc
   Reset {} -> emptyDoc
   Enable {} -> emptyDoc
   Bit      -> emptyDoc
   Bool     -> emptyDoc
   FileType -> emptyDoc
   Annotated _ ty -> verilogType ty
+  BiDirectional _ ty -> verilogType ty
   _        -> brackets (int (typeSize t -1) <> colon <> int 0)
 
 sigDecl :: VerilogM Doc -> HWType -> VerilogM Doc
@@ -342,31 +410,32 @@ decls ds = do
 
 -- | Add attribute notation to given declaration
 addAttrs
-  :: [Attr']
+  :: [Attr TextS.Text]
   -> VerilogM Doc
   -> VerilogM Doc
 addAttrs []     t = t
 addAttrs attrs' t =
   "(*" <+> attrs'' <+> "*)" <+> t
  where
-  attrs'' = string $ Text.intercalate ", " (map renderAttr attrs')
+  attrs'' = stringS $ TextS.intercalate ", " (map renderAttr attrs')
 
 -- | Convert single attribute to verilog syntax
-renderAttr :: Attr' -> Text.Text
-renderAttr (StringAttr'  key value) = pack $ concat [key, " = ", show value]
-renderAttr (IntegerAttr' key value) = pack $ concat [key, " = ", show value]
-renderAttr (BoolAttr'    key True ) = pack $ concat [key, " = ", "1"]
-renderAttr (BoolAttr'    key False) = pack $ concat [key, " = ", "0"]
-renderAttr (Attr'        key      ) = pack $ key
+renderAttr :: Attr TextS.Text -> TextS.Text
+renderAttr (StringAttr  key value) = TextS.concat [key, " = ", TextS.pack (show value)]
+renderAttr (IntegerAttr key value) = TextS.concat [key, " = ", TextS.pack (show value)]
+renderAttr (BoolAttr    key True ) = TextS.concat [key, " = ", "1"]
+renderAttr (BoolAttr    key False) = TextS.concat [key, " = ", "0"]
+renderAttr (Attr        key      ) = key
 
 decl :: Declaration -> VerilogM (Maybe Doc)
-decl (NetDecl' noteM wr id_ tyE iEM) =
-  Just A.<$> maybe id addNote noteM (addAttrs attrs (wireRegFileDoc wr tyE <+> tyDec tyE))
+decl (NetDecl' noteM id_ tyE iEM) = do
+  us <- use usages
+  let u = lookupUsage id_ us
+  Just A.<$> maybe id addNote noteM (addAttrs attrs (usageFileDoc u tyE <+> tyDec tyE))
   where
-    tyDec (Left  ty) = stringS ty <+> pretty id_ <> iE
-    tyDec (Right ty) = sigDecl (pretty id_) ty <> iE
+    tyDec ty = sigDecl (pretty id_) ty <> iE
     addNote n = mappend ("//" <+> stringS n <> line)
-    attrs = fromMaybe [] (hwTypeAttrs A.<$> either (const Nothing) Just tyE)
+    attrs = fromMaybe [] (hwTypeAttrs A.<$> Just tyE)
     iE    = maybe emptyDoc (noEmptyInit . expr_ False) iEM
 
 decl _ = return Nothing
@@ -471,7 +540,8 @@ inst_' id_ scrut scrutTy es = fmap Just $
 -- | Turn a Netlist Declaration to a Verilog concurrent block
 inst_ :: Declaration -> VerilogM (Maybe Doc)
 inst_ (TickDecl {}) = return Nothing
-inst_ (Assignment id_ e) = fmap Just $
+inst_ (CompDecl {}) = return Nothing
+inst_ (Assignment id_ Cont e) = fmap Just $
   "assign" <+> pretty id_ <+> equals <+> expr_ False e <> semi
 
 inst_ (CondAssignment id_ _ scrut _ [(Just (BoolLit b), l),(_,r)]) = fmap Just $
@@ -535,6 +605,9 @@ inst_ (NetDecl' {}) = return Nothing
 inst_ (ConditionalDecl cond ds) = Just <$>
   "`ifdef" <+> pretty cond <> line <> indent 2 (insts ds) <> line <> "`endif"
 
+inst_ d =
+  error ("inst_: " ++ show d)
+
 seq_ :: Seq -> VerilogM Doc
 seq_ (AlwaysClocked edge clk ds) =
   "always @" <>
@@ -575,8 +648,9 @@ seq_ (Branch scrut scrutTy es) =
       "end") <:> conds es'
 
 seq_ (SeqDecl sd) = case sd of
-  Assignment id_ e ->
-    pretty id_ <+> equals <+> expr_ False e <> semi
+  Assignment id_ (Proc b) e ->
+    let op = case b of { Blocking -> equals; NonBlocking -> "<=" }
+     in pretty id_ <+> op <+> expr_ False e <> semi
 
   BlackBoxD {} ->
     fromMaybe <$> emptyDoc <*> inst_ sd
@@ -970,7 +1044,7 @@ expr_ _ (Identifier id_ (Just (Indexed ((Unsigned w),_,_))))  = do
 expr_ _ (Identifier _ (Just (Indexed ((BitVector _),_,0)))) = do
   iw <- Ap $ use intWidth
   traceIf True ($(curLoc) ++ "WARNING: synthesizing bitvector mask to dontcare") $
-    verilogTypeErrValue (Signed iw)
+    verilogTypeErrValue (Unsigned iw)
 
 -- See [Note] bitvector projection
 expr_ _ (Identifier id_ (Just (Indexed ((BitVector w),_,1)))) = do
@@ -1060,16 +1134,16 @@ expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
 expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
   | pNm == "Clash.Sized.Internal.BitVector.fromInteger#"
   , [Literal _ (NumLit n), Literal _ m, Literal _ i] <- extractLiterals bbCtx
-  = let NumLit m' = m
-        NumLit i' = i
-    in exprLitV (Just (BitVector (fromInteger n),fromInteger n)) (BitVecLit m' i')
+  , NumLit m' <- m
+  , NumLit i' <- i
+  = exprLitV (Just (BitVector (fromInteger n),fromInteger n)) (BitVecLit m' i')
 
 expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
   | pNm == "Clash.Sized.Internal.BitVector.fromInteger##"
   , [Literal _ m, Literal _ i] <- extractLiterals bbCtx
-  = let NumLit m' = m
-        NumLit i' = i
-    in exprLitV (Just (Bit,1)) (BitLit $ toBit m' i')
+  , NumLit m' <- m
+  , NumLit i' <- i
+  = exprLitV (Just (Bit,1)) (BitLit $ toBit m' i')
 
 expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
   | pNm == "Clash.Sized.Internal.Index.fromInteger#"
@@ -1192,10 +1266,9 @@ bit_char k b = do
   udf <- Ap (use k)
   case (udf,b) of
     (Just Nothing,U)  -> char '0'
-    (Just (Just i),U) -> "'" <> int i <> "'"
+    (Just (Just i),U) -> int i
     _                 -> char (bit_char' b)
 
-
 dcToExpr :: HWType -> Int -> Expr
 dcToExpr ty i = Literal (Just (ty,conSize ty)) (NumLit (toInteger i))
 
@@ -1211,6 +1284,7 @@ punctuate' s d = vcat (punctuate s d) <> s
 
 encodingNote :: Applicative m => HWType -> m Doc
 encodingNote (Clock _) = string " // clock"
+encodingNote (ClockN _) = string " // clock (neg phase)"
 encodingNote (Reset _) = string " // reset"
 encodingNote (Enable _) = string " // enable"
 encodingNote (Annotated _ t) = encodingNote t
