@@ -38,6 +38,7 @@ import Clash.Pretty (ClashPretty(..), fromPretty, showDoc)
 whnf'
   :: Evaluator
   -> BindingMap
+  -> VarEnv Term
   -> TyConMap
   -> PrimHeap
   -> Supply
@@ -45,12 +46,12 @@ whnf'
   -> Bool
   -> Term
   -> (PrimHeap, PureHeap, Term)
-whnf' eval bm tcm ph ids is isSubj e =
+whnf' eval bm lh tcm ph ids is isSubj e =
   toResult $ whnf eval tcm isSubj m
  where
   toResult x = (mHeapPrim x, mHeapLocal x, mTerm x)
 
-  m  = Machine ph gh emptyVarEnv [] ids is e
+  m  = Machine ph gh lh [] ids is e
   gh = mapVarEnv bindingTerm bm
 
 -- | Evaluate to WHNF given an existing Heap and Stack
@@ -314,9 +315,19 @@ collectValueTicks = go []
 forcePrims :: Machine -> Bool
 forcePrims = go . mStack
  where
+  -- When do we need to force the compile-time evaluation of a primitive?
+  --
+  -- 1. When they are the subject of a case-expression
   go (Scrutinise{}:_) = True
+  -- 2. When they are in the argument position of another primitive:
+  --    primitives are assumed to be strict in their arguments
   go (PrimApply{}:_)  = True
+  -- We look through ticks
   go (Tickish{}:xs)   = go xs
+  -- We are in a context where we dereferenced a heap-binding, hence the
+  -- update fram on the stack. So now we need to check whether that variable
+  -- reference was in a position where the result must be evaluated to WHNF
+  go (Update{}:xs)    = go xs
   go _                = False
 
 primCount :: Machine -> Int
