@@ -86,8 +86,8 @@ import           Clash.Core.Term
 import           Clash.Core.Type           (Type (..))
 import           Clash.Core.VarEnv
 import           Clash.Core.Var            (Id, Var (..), TyVar, isGlobalId)
+import qualified Clash.Data.UniqMap as UniqMap
 import           Clash.Debug               (debugIsOn)
-import           Clash.Unique
 import           Clash.Util
 import           Clash.Pretty
 
@@ -457,7 +457,7 @@ checkValidSubst subst@(TvSubst inScope tenv) tys a =
        "needsInScope" <+> clashPretty needsInScope)
   a
  where
-  needsInScope = foldrWithUnique (\k _ s -> delVarSetByKey k s)
+  needsInScope = UniqMap.foldrWithUnique (\k _ s -> delVarSetByKey k s)
                    (freeVarsOf tys)
                    tenv
   tysFVsInSope = needsInScope `varSetInScope` inScope
@@ -968,10 +968,6 @@ acmpTerm' inScope = go (mkRnEnv inScope)
     Case {}    -> 11
     Tick {}    -> 12
 
-thenCompare :: Ordering -> Ordering -> Ordering
-thenCompare EQ rel = rel
-thenCompare rel _  = rel
-
 -- | Structural equality on 'Term'
 eqTerm :: Term -> Term -> Bool
 eqTerm = go
@@ -1040,5 +1036,3 @@ instance TypeError (
 
 instance Ord Term where
   compare = acmpTerm
-
-deriving instance Ord TickInfo
