@@ -65,17 +65,31 @@ import           Data.Hashable          (Hashable (hashWithSalt))
 import           Data.List              (foldl')
 import           Data.List.Extra        (splitAtList)
 import           Data.Maybe             (isJust, mapMaybe)
+import           Data.Text              (Text)
 import           GHC.Base               (isTrue#,(==#))
 import           GHC.Generics           (Generic(..))
 import           GHC.Integer            (smallInteger)
 import           GHC.Integer.Logarithms (integerLogBase#)
 import           GHC.TypeLits           (type TypeError, ErrorMessage(Text, (:<>:)))
+#if MIN_VERSION_base(4,16,0)
+import           GHC.Base               (ord)
+import           Data.Char              (chr)
+import           Data.Maybe             (fromMaybe)
+import           Data.Text.Extra        (showt)
+#endif
 
 -- GHC API
 #if MIN_VERSION_ghc(9,0,0)
+#if MIN_VERSION_ghc(9,2,0)
+import           GHC.Builtin.Names
+  (typeCharCmpTyFamNameKey, typeConsSymbolTyFamNameKey, typeUnconsSymbolTyFamNameKey,
+   typeCharToNatTyFamNameKey, typeNatToCharTyFamNameKey)
+#else
+import           GHC.Builtin.Names      (typeNatLeqTyFamNameKey)
+#endif
 import           GHC.Builtin.Names
   (integerTyConKey, typeNatAddTyFamNameKey, typeNatExpTyFamNameKey,
-   typeNatLeqTyFamNameKey, typeNatMulTyFamNameKey, typeNatSubTyFamNameKey,
+   typeNatMulTyFamNameKey, typeNatSubTyFamNameKey,
    typeNatCmpTyFamNameKey, ordLTDataConKey, ordEQDataConKey, ordGTDataConKey,
    typeSymbolAppendFamNameKey, typeSymbolCmpTyFamNameKey)
 import           GHC.Types.SrcLoc       (wiredInSrcSpan)
@@ -99,12 +113,13 @@ import           Unique                 (getKey)
 #endif
 
 -- Local imports
+import           Clash.Annotations.SynthesisAttributes
 import           Clash.Core.DataCon
 import           Clash.Core.Name
 import {-# SOURCE #-} Clash.Core.Subst
 import           Clash.Core.TyCon
 import           Clash.Core.Var
-import           Clash.Unique
+import qualified Clash.Data.UniqMap as UniqMap
 import           Clash.Util
 
 #if __GLASGOW_HASKELL__ <= 806
@@ -114,7 +129,7 @@ ordEQDataConKey = eqDataConKey
 ordGTDataConKey = gtDataConKey
 #endif
 
-varAttrs :: Var a -> [Attr']
+varAttrs :: Var a -> [Attr Text]
 varAttrs t@(TyVar {}) =
   error $ $(curLoc) ++ "Unexpected argument: " ++ show t
 
@@ -131,7 +146,7 @@ data Type
   | ForAllTy !TyVar !Type       -- ^ Polymorphic Type
   | AppTy    !Type !Type        -- ^ Type Application
   | LitTy    !LitTy             -- ^ Type literal
-  | AnnType  [Attr'] !Type      -- ^ Annotated type, see Clash.Annotations.SynthesisAttributes
+  | AnnType  [Attr Text] !Type  -- ^ Annotated type, see Clash.Annotations.SynthesisAttributes
   deriving (Show, Generic, NFData, Binary)
 
 instance TypeError (
@@ -158,6 +173,7 @@ data ConstTy
 data LitTy
   = NumTy !Integer
   | SymTy !String
+  | CharTy !Char
   deriving (Eq,Ord,Show,Generic,NFData,Hashable,Binary)
 
 -- | The level above types
@@ -263,7 +279,7 @@ coreView1 tcMap ty = case tyView ty of
     , [_,elTy] <- args
     -> Just elTy
     | otherwise
-    -> case tcMap `lookupUniqMap'` tcNm of
+    -> case UniqMap.find tcNm tcMap of
          AlgTyCon {algTcRhs = (NewTyCon _ nt)}
            -> newTyConInstRhs nt args
          _ -> reduceTypeFamily tcMap ty
@@ -373,9 +389,7 @@ isPolyFunCoreTy _ ty = case tyView ty of
 
 -- | Extract attributes from type. Will return an empty list if this is an
 -- AnnType with an empty list AND if this is not an AnnType at all.
-typeAttrs
-  :: Type
-  -> [Attr']
+typeAttrs :: Type -> [Attr Text]
 typeAttrs (AnnType attrs _typ) = attrs
 typeAttrs _                    = []
 
@@ -522,16 +536,18 @@ reduceTypeFamily tcm (tyView -> TyConApp tc tys)
         -> Just (LitTy (NumTy z))
       _ -> Nothing
 
+#if !MIN_VERSION_ghc(9,2,0)
   | nameUniq tc == getKey typeNatLeqTyFamNameKey
   = case mapMaybe (litView tcm) tys of
       [i1, i2]
-        | Just (FunTyCon {tyConKind = tck}) <- lookupUniqMap tc tcm
+        | Just (FunTyCon {tyConKind = tck}) <- UniqMap.lookup tc tcm
         , (_,tyView -> TyConApp boolTcNm []) <- splitFunTys tcm tck
-        , Just boolTc <- lookupUniqMap boolTcNm tcm
+        , Just boolTc <- UniqMap.lookup boolTcNm tcm
         -> let [falseTc,trueTc] = map (coerce . dcName) (tyConDataCons boolTc)
             in  if i1 <= i2 then Just (mkTyConApp trueTc [])
                             else Just (mkTyConApp falseTc [])
       _ -> Nothing
+#endif
 
   | nameUniq tc == getKey typeNatCmpTyFamNameKey -- "GHC.TypeNats.CmpNat"
   = case mapMaybe (litView tcm) tys of
@@ -553,6 +569,50 @@ reduceTypeFamily tcm (tyView -> TyConApp tc tys)
             GT -> Name User "GHC.Types.GT" (getKey ordGTDataConKey) wiredInSrcSpan
       _ -> Nothing
 
+#if MIN_VERSION_base(4,16,0)
+  | nameUniq tc == getKey typeCharCmpTyFamNameKey -- "GHC.TypeNats.CmpSymbol"
+  = case mapMaybe (charLitView tcm) tys of
+      [s1, s2] ->
+        Just $ ConstTy $ TyCon $
+          case compare s1 s2 of
+            LT -> Name User (showt 'LT) (getKey ordLTDataConKey) wiredInSrcSpan
+            EQ -> Name User (showt 'EQ) (getKey ordEQDataConKey) wiredInSrcSpan
+            GT -> Name User (showt 'GT) (getKey ordGTDataConKey) wiredInSrcSpan
+      _ -> Nothing
+
+  | nameUniq tc == getKey typeConsSymbolTyFamNameKey -- ConsSymbol
+  , [c0, s0] <- tys
+  , Just c1 <- charLitView tcm c0
+  , Just s1 <- symLitView tcm s0
+  = Just (LitTy (SymTy (c1:s1)))
+
+  | nameUniq tc == getKey typeUnconsSymbolTyFamNameKey -- UnconsSymbol
+  , [s1] <- mapMaybe (symLitView tcm) tys
+  = fromMaybe (error "reduceTypeFamily: cannot construct UnconsSymbol result") $ do
+      FunTyCon {tyConKind = tck} <- UniqMap.lookup tc tcm
+      TyConApp maybeTcNm [tupTcApp] <- pure (tyView (snd (splitFunTys tcm tck)))
+      maybeTc <- UniqMap.lookup maybeTcNm tcm
+      [nothingTc,justTc] <- pure (map (coerce . dcName) (tyConDataCons maybeTc))
+      TyConApp tupTcNm [charTy,symbolTy] <- pure (tyView tupTcApp)
+      tupTc <- UniqMap.lookup tupTcNm tcm
+      [tupDc] <- pure (map (coerce . dcName) (tyConDataCons tupTc))
+      case s1 of
+        [] ->
+          pure (Just (mkTyConApp nothingTc [tupTcApp]))
+        (c:cs) ->
+          let tup = mkTyConApp tupDc
+                      [charTy,symbolTy,LitTy (CharTy c),LitTy (SymTy cs)]
+           in pure (Just (mkTyConApp justTc [tupTcApp,tup]))
+
+  | nameUniq tc == getKey typeCharToNatTyFamNameKey -- CharToNat
+  , [c1] <- mapMaybe (charLitView tcm) tys
+  = Just (LitTy (NumTy (fromIntegral (ord c1))))
+
+  | nameUniq tc == getKey typeNatToCharTyFamNameKey -- NatToChar
+  , [n1] <- mapMaybe (litView tcm) tys
+  = Just (LitTy (CharTy (chr (fromInteger n1))))
+#endif
+
   | nameUniq tc == getKey typeSymbolAppendFamNameKey  -- GHC.TypeLits.AppendSymbol"
   = case mapMaybe (symLitView tcm) tys of
       [s1, s2] ->
@@ -614,7 +674,7 @@ reduceTypeFamily tcm (tyView -> TyConApp tc tys)
         -> Just (LitTy (NumTy (i1 `mod` i2)))
       _ -> Nothing
 
-  | Just (FunTyCon {tyConSubst = tcSubst}) <- lookupUniqMap tc tcm
+  | Just (FunTyCon {tyConSubst = tcSubst}) <- UniqMap.lookup tc tcm
   = let -- See [Note: Eager type families]
         tysR = map (argView tcm) tys
      in findFunSubst tcm tcSubst tysR
@@ -624,7 +684,7 @@ reduceTypeFamily _ _ = Nothing
 -- |
 isTypeFamilyApplication ::  TyConMap -> Type -> Bool
 isTypeFamilyApplication tcm (tyView -> TyConApp tcNm _args)
-  | Just (FunTyCon {}) <- lookupUniqMap tcNm tcm = True
+  | Just (FunTyCon {}) <- UniqMap.lookup tcNm tcm = True
 isTypeFamilyApplication _tcm _type = False
 
 argView :: TyConMap -> Type -> Type
@@ -642,6 +702,13 @@ symLitView _ (LitTy (SymTy s))                = Just s
 symLitView m (reduceTypeFamily m -> Just ty') = symLitView m ty'
 symLitView _ _ = Nothing
 
+#if MIN_VERSION_base(4,16,0)
+charLitView :: TyConMap -> Type -> Maybe Char
+charLitView _ (LitTy (CharTy c)) = Just c
+charLitView m (reduceTypeFamily m -> Just t) = charLitView m t
+charLitView _ _ = Nothing
+#endif
+
 isIntegerTy :: Type -> Bool
 isIntegerTy (ConstTy (TyCon nm)) = nameUniq nm == getKey integerTyConKey
 isIntegerTy _ = False
@@ -663,7 +730,7 @@ normalizeType tcMap = go
         nameOcc tcNm == "Clash.Sized.Internal.Unsigned.Unsigned"
       -> mkTyConApp tcNm (map go args)
       | otherwise
-      -> case lookupUniqMap' tcMap tcNm of
+      -> case UniqMap.find tcNm tcMap of
           AlgTyCon {algTcRhs = (NewTyCon _ nt)}
              -> case newTyConInstRhs nt args of
                   Just ty' -> go ty'
@@ -691,7 +758,7 @@ isClassTy
   -> Type
   -> Bool
 isClassTy tcm (tyView -> TyConApp tcNm _) =
-  case lookupUniqMap tcNm tcm of
+  case UniqMap.lookup tcNm tcm of
     Just (AlgTyCon {isClassTc}) -> isClassTc
     _ -> False
 isClassTy _ _ = False
