@@ -1,6 +1,6 @@
 {-|
   Copyright   :  (C) 2012-2016, University of Twente,
-                     2021     , QBayLogic B.V.,
+                     2021-2023, QBayLogic B.V.,
                      2022     , Google Inc.
   License     :  BSD2 (see the file LICENSE)
   Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
@@ -20,13 +20,14 @@ module Clash.Core.Util where
 
 import           Control.Concurrent.Supply     (Supply, freshId)
 import Control.Monad.Trans.Except              (Except, throwE, runExcept)
-import Data.Bifunctor                          (first)
+import Data.Bifunctor                          (first, second)
 import qualified Data.HashSet                  as HashSet
 import qualified Data.Graph                    as Graph
-import Data.List                               (mapAccumR)
+import Data.List                               (mapAccumR, uncons)
 import Data.List.Extra                         (zipEqual)
+import Data.List.NonEmpty                      (NonEmpty (..), toList)
 import Data.Maybe
-  (fromJust, isJust, mapMaybe, catMaybes)
+  (fromMaybe, isJust, mapMaybe, catMaybes)
 import qualified Data.Set                      as Set
 import qualified Data.Set.Lens                 as Lens
 import qualified Data.Text                     as T
@@ -55,8 +56,8 @@ import Clash.Core.Type
 import Clash.Core.TysPrim                (liftedTypeKind, typeNatKind)
 import Clash.Core.Var                    (Id, Var(..), mkLocalId, mkTyVar)
 import Clash.Core.VarEnv
+import qualified Clash.Data.UniqMap as UniqMap
 import Clash.Debug                       (traceIf)
-import Clash.Unique
 import Clash.Util
 
 -- | Rebuild a let expression / let expressions by taking the SCCs of a list
@@ -105,12 +106,15 @@ mkVec nilCon consCon resTy = go
                                         ,Left x
                                         ,Left (go (n-1) xs)]
 
-    nilCoTy    = head (fromJust $! dataConInstArgTys nilCon  [(LitTy (NumTy 0))
-                                                             ,resTy])
-    consCoTy n = head (fromJust $! dataConInstArgTys consCon
-                                                     [(LitTy (NumTy n))
-                                                     ,resTy
-                                                     ,(LitTy (NumTy (n-1)))])
+    nilCoTy    = case dataConInstArgTys nilCon [(LitTy (NumTy 0)) ,resTy] of
+                   Just (x:_) -> x
+                   _ -> error "impossible"
+    consCoTy n = case dataConInstArgTys consCon
+                                        [(LitTy (NumTy n))
+                                        ,resTy
+                                        ,(LitTy (NumTy (n-1)))] of
+                   Just (x:_) -> x
+                   _ -> error "impossible"
 
 -- | Append elements to the supplied vector
 appendToVec :: DataCon -- ^ The Cons (:>) constructor
@@ -129,16 +133,19 @@ appendToVec consCon resTy vec = go
                                         ,Left x
                                         ,Left (go (n-1) xs)]
 
-    consCoTy n = head (fromJust $! dataConInstArgTys consCon
-                                                   [(LitTy (NumTy n))
-                                                   ,resTy
-                                                   ,(LitTy (NumTy (n-1)))])
+    consCoTy n = case dataConInstArgTys consCon
+                                        [(LitTy (NumTy n))
+                                        ,resTy
+                                        ,(LitTy (NumTy (n-1)))] of
+                   Just (x:_) -> x
+                   _ -> error "impossible"
 
 -- | Create let-bindings with case-statements that select elements out of a
 -- vector. Returns both the variables to which element-selections are bound
 -- and the let-bindings
 extractElems
-  :: Supply
+  :: HasCallStack
+  => Supply
   -- ^ Unique supply
   -> InScopeSet
   -- ^ (Superset of) in scope variables
@@ -152,22 +159,22 @@ extractElems
   -- ^ Length of the vector
   -> Term
   -- ^ The vector
-  -> (Supply, [(Term,[(Id, Term)])])
+  -> (Supply, NonEmpty (Term,NonEmpty (Id, Term)))
 extractElems supply inScope consCon resTy s maxN vec =
-  first fst (go maxN (supply,inScope) vec)
+  if maxN >= 1 then
+    first fst (go maxN (supply,inScope) vec)
+  else
+    error "extractElems must be called with positive number"
  where
   go :: Integer -> (Supply,InScopeSet) -> Term
-     -> ((Supply,InScopeSet),[(Term,[(Id, Term)])])
-  go 0 uniqs _ = (uniqs,[])
-  go n uniqs0 e =
-    (uniqs3,(elNVar,[(elNId, lhs),(restNId, rhs)]):restVs)
-   where
-    tys = [(LitTy (NumTy n)),resTy,(LitTy (NumTy (n-1)))]
-    (Just idTys) = dataConInstArgTys consCon tys
-    restTy       = last idTys
+     -> ((Supply,InScopeSet),NonEmpty (Term, NonEmpty (Id, Term)))
+  go n uniqs0 e = fromMaybe (error "extractElems: failed to project elements") $ do
+    let tys = [(LitTy (NumTy n)),resTy,(LitTy (NumTy (n-1)))]
+    idTys <- dataConInstArgTys consCon tys
+    let restTy = last idTys
 
-    (uniqs1,mTV) = mkUniqSystemTyVar uniqs0 ("m",typeNatKind)
-    (uniqs2,[elNId,restNId,co,el,rest]) =
+    let (uniqs1,mTV) = mkUniqSystemTyVar uniqs0 ("m",typeNatKind)
+    (uniqs2,[elNId,restNId,co,el,rest]) <- pure $
       mapAccumR mkUniqSystemId uniqs1 $ zipEqual
         ["el" `T.append` (s `T.cons` T.pack (show (maxN-n)))
         ,"rest" `T.append` (s `T.cons` T.pack (show (maxN-n)))
@@ -177,12 +184,14 @@ extractElems supply inScope consCon resTy s maxN vec =
         ]
         (resTy:restTy:idTys)
 
-    elNVar    = Var elNId
-    pat       = DataPat consCon [mTV] [co,el,rest]
-    lhs       = Case e resTy  [(pat,Var el)]
-    rhs       = Case e restTy [(pat,Var rest)]
+    let elNVar    = Var elNId
+        pat       = DataPat consCon [mTV] [co,el,rest]
+        lhs       = Case e resTy  [(pat,Var el)]
+        rhs       = Case e restTy [(pat,Var rest)]
 
-    (uniqs3,restVs) = go (n-1) uniqs2 (Var restNId)
+    let (uniqs3,restVs) = if n < 2 then (uniqs2,[]) else second toList (go (n-1) uniqs2 (Var restNId))
+
+    return (uniqs3,(elNVar,(elNId, lhs) :| [(restNId, rhs)]) :| restVs)
 
 -- | Create let-bindings with case-statements that select elements out of a
 -- tree. Returns both the variables to which element-selections are bound
@@ -214,35 +223,31 @@ extractTElems supply inScope lrCon brCon resTy s maxN tree =
      -> (Supply,InScopeSet)
      -> Term
      -> ((Supply,InScopeSet),([Term],[(Id, Term)]))
-  go 0 _ ks uniqs0 e = (uniqs1,([elNVar],[(elNId, rhs)]))
-   where
-    tys          = [LitTy (NumTy 0),resTy]
-    (Just idTys) = dataConInstArgTys lrCon tys
+  go 0 _ ks uniqs0 e = fromMaybe (error "extractTElems: failed to project elements") $ do
+    let tys = [LitTy (NumTy 0),resTy]
+    idTys <- dataConInstArgTys lrCon tys
+    (k,_) <- uncons ks
 
-    (uniqs1,[elNId,co,el]) =
+    (uniqs1,[elNId,co,el]) <- pure $
       mapAccumR mkUniqSystemId uniqs0 $ zipEqual
-        [ "el" `T.append` (s `T.cons` T.pack (show (head ks)))
+        [ "el" `T.append` (s `T.cons` T.pack (show k))
         , "_co_"
         , "el"
         ]
         (resTy:idTys)
-    elNVar = Var elNId
-    pat    = DataPat lrCon [] [co,el]
-    rhs    = Case e resTy [(pat,Var el)]
+    let elNVar = Var elNId
+        pat    = DataPat lrCon [] [co,el]
+        rhs    = Case e resTy [(pat,Var el)]
+    return (uniqs1,([elNVar],[(elNId, rhs)]))
 
-  go n bs ks uniqs0 e =
-    (uniqs4
-    ,(lVars ++ rVars,(ltNId, ltRhs):
-                     (rtNId, rtRhs):
-                     (lBinds ++ rBinds)))
-   where
-    tys = [LitTy (NumTy n),resTy,LitTy (NumTy (n-1))]
-    (Just idTys) = dataConInstArgTys brCon tys
+  go n bs ks uniqs0 e = fromMaybe (error "extractTElems: failed to project elements") $ do
+    let tys = [LitTy (NumTy n),resTy,LitTy (NumTy (n-1))]
+    idTys <- dataConInstArgTys brCon tys
 
-    (uniqs1,mTV) = mkUniqSystemTyVar uniqs0 ("m",typeNatKind)
-    (b0:bL,b1:bR) = splitAt (length bs `div` 2) bs
-    brTy = last idTys
-    (uniqs2,[ltNId,rtNId,co,lt,rt]) =
+    let (uniqs1,mTV) = mkUniqSystemTyVar uniqs0 ("m",typeNatKind)
+    (b0:bL,b1:bR) <- pure (splitAt (length bs `div` 2) bs)
+    let brTy = last idTys
+    (uniqs2,[ltNId,rtNId,co,lt,rt]) <- pure $
       mapAccumR mkUniqSystemId uniqs1 $ zipEqual
         ["lt" `T.append` (s `T.cons` T.pack (show b0))
         ,"rt" `T.append` (s `T.cons` T.pack (show b1))
@@ -251,15 +256,21 @@ extractTElems supply inScope lrCon brCon resTy s maxN tree =
         ,"rt"
         ]
         (brTy:brTy:idTys)
-    ltVar = Var ltNId
-    rtVar = Var rtNId
-    pat   = DataPat brCon [mTV] [co,lt,rt]
-    ltRhs = Case e brTy [(pat,Var lt)]
-    rtRhs = Case e brTy [(pat,Var rt)]
+    let ltVar = Var ltNId
+        rtVar = Var rtNId
+        pat   = DataPat brCon [mTV] [co,lt,rt]
+        ltRhs = Case e brTy [(pat,Var lt)]
+        rtRhs = Case e brTy [(pat,Var rt)]
 
-    (kL,kR) = splitAt (length ks `div` 2) ks
-    (uniqs3,(lVars,lBinds)) = go (n-1) bL kL uniqs2 ltVar
-    (uniqs4,(rVars,rBinds)) = go (n-1) bR kR uniqs3 rtVar
+        (kL,kR) = splitAt (length ks `div` 2) ks
+        (uniqs3,(lVars,lBinds)) = go (n-1) bL kL uniqs2 ltVar
+        (uniqs4,(rVars,rBinds)) = go (n-1) bR kR uniqs3 rtVar
+
+    return ( uniqs4
+           , ( lVars ++ rVars
+             , (ltNId, ltRhs):(rtNId, rtRhs): (lBinds ++ rBinds)
+             )
+           )
 
 -- | Create a vector of supplied elements
 mkRTree :: DataCon -- ^ The LR constructor
@@ -285,12 +296,15 @@ mkRTree lrCon brCon resTy = go
                               ,Left (go (n-1) xsL)
                               ,Left (go (n-1) xsR)]
 
-    lrCoTy   = head (fromJust $! dataConInstArgTys lrCon  [(LitTy (NumTy 0))
-                                                         ,resTy])
-    brCoTy n = head (fromJust $! dataConInstArgTys brCon
-                                                   [(LitTy (NumTy n))
-                                                   ,resTy
-                                                   ,(LitTy (NumTy (n-1)))])
+    lrCoTy   = case dataConInstArgTys lrCon [(LitTy (NumTy 0)) ,resTy] of
+                 Just (x:_) -> x
+                 _ -> error "impossible"
+    brCoTy n = case dataConInstArgTys brCon
+                                      [(LitTy (NumTy n))
+                                      ,resTy
+                                      ,(LitTy (NumTy (n-1)))] of
+                 Just (x:_) -> x
+                 _ -> error "impossible"
 
 -- | Determine whether a type is isomorphic to "Clash.Signal.Internal.Signal"
 --
@@ -315,7 +329,7 @@ isSignalType tcm ty = go HashSet.empty ty
       "Clash.Signal.BiSignal.BiSignalIn"  -> True
       "Clash.Signal.BiSignal.BiSignalOut" -> True
       _ | tcNm `HashSet.member` tcSeen    -> False -- Do not follow rec types
-        | otherwise -> case lookupUniqMap tcNm tcm of
+        | otherwise -> case UniqMap.lookup tcNm tcm of
             Just tc -> let dcs         = tyConDataCons tc
                            dcInsArgTys = concat
                                        $ mapMaybe (`dataConInstArgTys` args) dcs
@@ -344,6 +358,7 @@ isClockOrReset
 isClockOrReset m (coreView1 m -> Just ty)    = isClockOrReset m ty
 isClockOrReset _ (tyView -> TyConApp tcNm _) = case nameOcc tcNm of
   "Clash.Signal.Internal.Clock" -> True
+  "Clash.Signal.Internal.ClockN" -> True
   "Clash.Signal.Internal.Reset" -> True
   _ -> False
 isClockOrReset _ _ = False
@@ -472,6 +487,7 @@ undefinedPrims =
   , "GHC.Err.error"
   , "GHC.Err.errorWithoutStackTrace"
   , "GHC.Err.undefined"
+  , "GHC.Prim.Panic.absentError"
   , "GHC.Real.divZeroError"
   , "GHC.Real.overflowError"
   , "GHC.Real.ratioZeroDenominatorError"
@@ -552,7 +568,7 @@ shouldSplit0
   -> TypeView
   -> Maybe ([Term] -> Term, Projections, [Type])
 shouldSplit0 tcm (TyConApp tcNm tyArgs)
-  | Just tc <- lookupUniqMap tcNm tcm
+  | Just tc <- UniqMap.lookup tcNm tcm
   , [dc] <- tyConDataCons tc
   , let dcArgs = substArgTys dc tyArgs
   , let dcArgsLen = length dcArgs
@@ -571,7 +587,7 @@ shouldSplit0 tcm (TyConApp tcNm tyArgs)
   , [nTy,argTy] <- tyArgs
   , Right n <- runExcept (tyNatSize tcm nTy)
   , n > 1
-  , Just tc <- lookupUniqMap tcNm tcm
+  , Just tc <- UniqMap.lookup tcNm tcm
   , [nil,cons] <- tyConDataCons tc
   = if shouldSplitTy (tyView (coreView tcm argTy)) then
       Just ( mkVec nil cons argTy n
@@ -603,7 +619,7 @@ shouldSplit0 tcm (TyConApp tcNm tyArgs)
   -- to it being a zero-width type
   --
   -- TODO: This currently only handles (IP $x, KnownDomain) given that $x is any
-  -- TODO: of the constructs handled in 'splitTy'. In practise this means only
+  -- TODO: of the constructs handled in 'splitTy'. In practice this means only
   -- TODO: HiddenClock, HiddenReset, and HiddenEnable are handled. If a user were
   -- TODO: to define their own versions with -for example- the elements of the
   -- TODO: tuple swapped, 'isHidden' wouldn't recognize it. We could generalize
@@ -616,9 +632,9 @@ shouldSplit0 tcm (TyConApp tcNm tyArgs)
     && nameOcc a2Nm == "Clash.Signal.Internal.KnownDomain"
   isHidden _ _ = False
 
-  -- Currently we're only interested in splitting of Clock, Reset, and Enable
   splitTy (TyConApp tcNm0 _)
     = nameOcc tcNm0 `elem` [ "Clash.Signal.Internal.Clock"
+                           , "Clash.Signal.Internal.ClockN"
                            , "Clash.Signal.Internal.Reset"
                            , "Clash.Signal.Internal.Enable"
                            -- iverilog doesn't like it when we put file handles
@@ -663,9 +679,9 @@ stripIP t = t
 -- | Do an inverse topological sorting of the let-bindings in a let-expression
 inverseTopSortLetBindings
   :: HasCallStack
-  => Term
-  -> Term
-inverseTopSortLetBindings (Letrec bndrs0 res) =
+  => [(Id, Term)]
+  -> [(Id, Term)]
+inverseTopSortLetBindings bndrs0 =
   let (graph,nodeMap,_) =
         Graph.graphFromEdges
           (map (\(i,e) -> let fvs = fmap varUniq
@@ -673,7 +689,7 @@ inverseTopSortLetBindings (Letrec bndrs0 res) =
                           in  ((i,e),varUniq i,fvs)) bndrs0)
       nodes  = postOrd graph
       bndrs1 = map ((\(x,_,_) -> x) . nodeMap) nodes
-  in  Letrec bndrs1 res
+   in bndrs1
  where
   postOrd :: Graph.Graph -> [Graph.Vertex]
   postOrd g = postorderF (Graph.dff g) []
@@ -683,8 +699,6 @@ inverseTopSortLetBindings (Letrec bndrs0 res) =
 
   postorder :: Graph.Tree a -> [a] -> [a]
   postorder (Graph.Node a ts) = postorderF ts . (a :)
-
-inverseTopSortLetBindings e = e
 {-# SCC inverseTopSortLetBindings #-}
 
 -- | Group let-bindings into cyclic groups and acyclic individual bindings
@@ -714,12 +728,14 @@ mkSelectorCase caller inScope tcm scrut dcI fieldI = go (inferCoreTypeOf tcm scr
   where
     go (coreView1 tcm -> Just ty') = go ty'
     go scrutTy@(tyView -> TyConApp tc args) =
-      case tyConDataCons (lookupUniqMap' tcm tc) of
+      case tyConDataCons (UniqMap.find tc tcm) of
         [] -> cantCreate $(curLoc) ("TyCon has no DataCons: " ++ show tc ++ " " ++ showPpr tc) scrutTy
         dcs | dcI > length dcs -> cantCreate $(curLoc) "DC index exceeds max" scrutTy
             | otherwise -> do
           let dc = indexNote ($(curLoc) ++ "No DC with tag: " ++ show (dcI-1)) dcs (dcI-1)
-          let (Just fieldTys) = dataConInstArgTysE inScope tcm dc args
+          let fieldTys =
+                fromMaybe (cantCreate $(curLoc) "Cannot instantiate dataCon" scrutTy)
+                          (dataConInstArgTysE inScope tcm dc args)
           if fieldI >= length fieldTys
             then cantCreate $(curLoc) "Field index exceed max" scrutTy
             else do
@@ -732,6 +748,7 @@ mkSelectorCase caller inScope tcm scrut dcI fieldI = go (inferCoreTypeOf tcm scr
               return retVal
     go scrutTy = cantCreate $(curLoc) ("Type of subject is not a datatype: " ++ showPpr scrutTy) scrutTy
 
+    cantCreate :: String -> String -> Type -> a
     cantCreate loc info scrutTy = error $ loc ++ "Can't create selector " ++ show (caller,dcI,fieldI) ++ " for: (" ++ showPpr scrut ++ " :: " ++ showPpr scrutTy ++ ")\nAdditional info: " ++ info
 
 -- | Make a binder that should not be referenced
