@@ -2,7 +2,7 @@
   Copyright   :  (C) 2012-2016, University of Twente,
                      2016-2017, Myrtle Software Ltd,
                      2017     , QBayLogic, Google Inc.
-                     2020-2022, QBayLogic,
+                     2020-2023, QBayLogic,
                      2022     , Google Inc.
 
   License     :  BSD2 (see the file LICENSE)
@@ -35,7 +35,7 @@ import           Control.Monad.State              (evalState, get)
 import           Control.Monad.State.Strict       (State)
 import qualified Control.Monad.State.Strict       as State
 import qualified Crypto.Hash.SHA256               as Sha256
-import           Data.Bifunctor                   (first)
+import           Data.Bifunctor                   (first, second)
 import           Data.ByteString                  (ByteString)
 import qualified Data.ByteString                  as ByteString
 import qualified Data.ByteString.Lazy             as ByteStringLazy
@@ -48,6 +48,7 @@ import qualified Data.HashMap.Strict              as HashMap
 import qualified Data.HashSet                     as HashSet
 import           Data.Proxy                       (Proxy(..))
 import           Data.List                        (intercalate)
+import qualified Data.List                        as List
 import           Data.Maybe                       (fromMaybe, maybeToList, mapMaybe)
 import qualified Data.Map.Ordered                 as OMap
 import           Data.Map.Ordered.Extra           ()
@@ -138,11 +139,12 @@ import qualified Clash.Primitives.Sized.Vector    as P
 import qualified Clash.Primitives.GHC.Int         as P
 import qualified Clash.Primitives.GHC.Word        as P
 import qualified Clash.Primitives.Intel.ClockGen  as P
+import qualified Clash.Primitives.Magic           as P
 import qualified Clash.Primitives.Verification    as P
+import qualified Clash.Primitives.Xilinx.ClockGen as P
 import           Clash.Primitives.Types
 import           Clash.Signal.Internal
 import           Clash.Unique                     (Unique, getUnique)
-import           Clash.Util.Interpolate           (i)
 import           Clash.Util
   (ClashException(..), reportTimeDiff,
    wantedLanguageExtensions, unwantedLanguageExtensions, curLoc)
@@ -187,7 +189,7 @@ splitTopAnn tcm sp typ@(tyView -> FunTy {}) t@Synthesize{t_inputs} =
             in
               go newLam (take n portNames1 ++ ps)
           PortName nm ->
-            throw (flip (ClashException sp) Nothing $ [i|
+            throw (flip (ClashException sp) Nothing $ [I.i|
               Couldn't separate clock, reset, or enable from a product type due
               to a malformed Synthesize annotation. All clocks, resets, and
               enables should be given a unique port name. Type to be split:
@@ -314,8 +316,8 @@ generateHDL
   -> WHNF.Evaluator
   -- ^ Hardcoded evaluator for WHNF (old evaluator)
   -> Maybe (TopEntityT, [TopEntityT])
-  -- ^ Main top entity to compile. If Nothing, all top entities in previous
-  -- argument will be compiled.
+  -- ^ Main top entity to compile. If Nothing, all top entities in the
+  -- 'ClashDesign' argument will be compiled.
   -> Clock.UTCTime
   -> IO ()
 generateHDL env design hdlState typeTrans peEval eval mainTopEntity startTime = do
@@ -362,7 +364,7 @@ generateHDL env design hdlState typeTrans peEval eval mainTopEntity startTime =
     -> TopEntityT
     -> IO ()
   go compNames seenV edamFilesV ioLockV deps topEntityMap (TopEntityT topEntity annM isTb) = do
-  let domainConfs = designDomains design
+  let domainConfs = envDomains env
   let bindingsMap = designBindings design
   let primMap = envPrimitives env
   let topEntities0 = designEntities design
@@ -468,7 +470,7 @@ generateHDL env design hdlState typeTrans peEval eval mainTopEntity startTime =
 
       -- 4. Generate topEntity wrapper
       (hdlDocs, dfiles, mfiles) <- withMVar seenV $ \seen ->
-        pure $! createHDL hdlState' modNameT seen netlist domainConfs topComponent topNmT
+        pure $! createHDL hdlState' opts modNameT seen netlist domainConfs topComponent topNmT
 
       -- TODO: Data files should go into their own directory
       -- FIXME: Files can silently overwrite each other
@@ -479,9 +481,11 @@ generateHDL env design hdlState typeTrans peEval eval mainTopEntity startTime =
       let
         components = map (snd . snd) (OMap.assocs netlist)
         filesAndDigests0 =
-             zip (map fst hdlDocs) hdlDocDigests
+          -- FIXME: We should track dependencies of `mfiles` and `dfiles` and
+          -- maintain the proper topological sort of all these.
+             zip (map fst mfiles) memoryFilesDigests
           <> zip (map fst dfiles) dataFilesDigests
-          <> zip (map fst mfiles) memoryFilesDigests
+          <> zip (map fst hdlDocs) hdlDocDigests
 
       filesAndDigests1 <- modifyMVar edamFilesV $ \edamFiles ->
         if opt_edalize opts
@@ -574,6 +578,7 @@ knownBlackBoxFunctions =
     [ ('P.checkBBF, P.checkBBF)
     , ('P.bvToIntegerVHDL, P.bvToIntegerVHDL)
     , ('P.bvToIntegerVerilog, P.bvToIntegerVerilog)
+    , ('P.clashCompileErrorBBF, P.clashCompileErrorBBF)
     , ('P.foldBBF, P.foldBBF)
     , ('P.indexIntVerilog, P.indexIntVerilog)
     , ('P.indexToIntegerVerilog, P.indexToIntegerVerilog)
@@ -597,6 +602,10 @@ knownTemplateFunctions =
     , ('P.alteraPllTF, P.alteraPllTF)
     , ('P.altpllTF, P.altpllTF)
     , ('P.fromIntegerTFvhdl, P.fromIntegerTFvhdl)
+    , ('P.clockWizardTF, P.clockWizardTF)
+    , ('P.clockWizardDifferentialTF, P.clockWizardDifferentialTF)
+    , ('P.clockWizardTclTF, P.clockWizardTclTF)
+    , ('P.clockWizardDifferentialTclTF, P.clockWizardDifferentialTclTF)
     ]
 
 -- | Compiles blackbox functions and parses blackbox templates.
@@ -662,14 +671,14 @@ compilePrimitive idirs pkgDbs topDir (BlackBoxHaskell bbName wf usedArgs multiRe
       loadImportAndInterpret idirs args topDir qualMod funcName "BlackBoxFunction"
 
 compilePrimitive idirs pkgDbs topDir
-  (BlackBox pNm wf rVoid multiRes tkind () oReg libM imps fPlural incs rM riM templ) = do
+  (BlackBox pNm wf rVoid multiRes tkind () outputUsage libM imps fPlural incs rM riM templ) = do
   libM'  <- mapM parseTempl libM
   imps'  <- mapM parseTempl imps
   incs'  <- mapM (traverse parseBB) incs
   templ' <- parseBB templ
   rM'    <- traverse parseBB rM
   riM'   <- traverse parseBB riM
-  return (BlackBox pNm wf rVoid multiRes tkind () oReg libM' imps' fPlural incs' rM' riM' templ')
+  return (BlackBox pNm wf rVoid multiRes tkind () outputUsage libM' imps' fPlural incs' rM' riM' templ')
  where
   iArgs = concatMap (("-package-db":) . (:[])) pkgDbs
 
@@ -747,6 +756,8 @@ createHDL
   :: Backend backend
   => backend
   -- ^ Backend
+  -> ClashOpts
+  -- ^ Global Clash options
   -> IdentifierText
   -- ^ Module hierarchy root
   -> Id.IdentifierSet
@@ -762,21 +773,21 @@ createHDL
   -> ([(String,Doc)],[(String,FilePath)],[(String,String)])
   -- ^ The pretty-printed HDL documents
   -- + The data files that need to be copied
-createHDL backend modName seen components domainConfs top topName = flip evalState backend $ getAp $ do
+createHDL backend opts modName seen components domainConfs top topName = flip evalState backend $ getAp $ do
   let componentsL = map snd (OMap.assocs components)
-  (hdlNmDocs,incs) <-
+  (hdlNmDocs0,incs) <-
     fmap unzip $
-      forM componentsL $ \(ComponentMeta{cmLoc, cmScope}, comp) ->
-         genHDL modName cmLoc (Id.union seen cmScope) comp
+      forM componentsL $ \(ComponentMeta{cmLoc, cmScope,cmUsage}, comp) ->
+         genHDL opts modName cmLoc (Id.union seen cmScope) cmUsage comp
 
   hwtys <- HashSet.toList <$> extractTypes <$> Ap get
-  typesPkg <- mkTyPackage modName hwtys
+  typesPkg0 <- mkTyPackage modName hwtys
   dataFiles <- Ap getDataFiles
   memFiles  <- Ap getMemoryDataFiles
   let
-    hdl = map (first (<.> Clash.Backend.extension backend)) (typesPkg ++ hdlNmDocs)
-    qincs = concat incs
-    topFiles = hdl ++ qincs
+    typesPkg1 = map (first (<.> Clash.Backend.extension backend)) typesPkg0
+    hdlNmDocs1 = map (first (<.> Clash.Backend.extension backend)) hdlNmDocs0
+    topFiles = concat incs ++ typesPkg1 ++ hdlNmDocs1
 
     topClks = findClocks top
     sdcInfo = fmap findDomainConfig <$> topClks
@@ -810,7 +821,7 @@ pprVerilatorShim (Id.toText -> topNm) =
   -- Extra newlines are aggressively inserted so the quasiquoter doesn't wrap
   -- the outlines lines in the file. It doesn't matter for code inside main,
   -- but is fatal for the #include directives.
-  pretty $ Data.Text.pack [i|
+  pretty $ Data.Text.pack [I.i|
     \#include <cstdlib>
 
     \#include <verilated.h>
@@ -1080,28 +1091,37 @@ normalizeEntity env bindingsMap typeTrans peEval eval topEntities supply tm = tr
                             typeTrans peEval eval emptyVarEnv
                             topEntities doNorm
 
--- | topologically sort the top entities
-sortTop
-  :: BindingMap
-  -> [TopEntityT]
-  -> ([TopEntityT], HashMap Unique [Unique])
+-- | Reverse topologically sort given top entities. Also returns a mapping that
+-- maps a top entity to its reverse topologically sorted transitive dependencies.
+sortTop ::
+  BindingMap ->
+  [TopEntityT] ->
+  ( [TopEntityT]
+  , HashMap Unique [Unique]
+  )
 sortTop bindingsMap topEntities =
-  let (nodes,edges) = unzip (map go topEntities)
-      edges' = concat edges
-  in  case reverseTopSort nodes edges' of
-        Left msg   -> error msg
-        Right tops -> (tops, mapFrom edges')
+  case reverseTopSort nodes edges of
+    Left msg   -> error msg
+    Right tops -> (tops, mapFrom tops)
  where
-  go t@(TopEntityT topE _ _) =
-    let topRefs = goRefs topE topE
-    in  ((varUniq topE,t)
-         ,map ((\top -> (varUniq topE, varUniq (topId top)))) topRefs)
+  nodes = [(varUniq topE, t) | t@(TopEntityT topE _ _) <- topEntities]
+  edges = concatMap getEdges topEntities
 
-  goRefs top i_ =
-    let cg = callGraph bindingsMap i_
+  getEdges (TopEntityT topE _ _) =
+    map
+      (\top -> (varUniq topE, topToUnique top))
+      (getTransitiveRefs topE)
+
+  getTransitiveRefs top =
+    let allDeps = callGraph bindingsMap top
+    in  filter (\t -> topId t /= top && topId t `elemVarEnv` allDeps) topEntities
+
+  topToUnique = varUniq . topId
+
+  mapFrom tops =
+    let
+      topIndices = HashMap.fromList (zip (map topToUnique tops) [(0 :: Int)..])
+      nonOrdered = HashMap.fromListWith (<>) (map (second pure) edges)
+      orderFunc k = fromMaybe (-1) (HashMap.lookup k topIndices)
     in
-      filter
-        (\t -> topId t /= top && topId t /= i_ && topId t `elemVarEnv` cg)
-        topEntities
-
-  mapFrom = HashMap.fromListWith mappend . fmap (fmap pure)
+      HashMap.map (List.sortOn orderFunc) nonOrdered
