@@ -1,5 +1,5 @@
 {-|
-Copyright  : (C) 2021-2022, QBayLogic B.V.
+Copyright  : (C) 2021-2023, QBayLogic B.V.
 License    : BSD2 (see the file LICENSE)
 Maintainer : QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -58,7 +58,7 @@ import           Clash.Driver.Types
 import           Clash.Primitives.Types
 import           Clash.Core.Var (Id, varName)
 import           Clash.Netlist.Types
-  (TopEntityT, Component(..), HWType (Clock), hwTypeDomain)
+  (TopEntityT, Component(..), HWType (Clock, ClockN), hwTypeDomain)
 import qualified Clash.Netlist.Types as Netlist
 import qualified Clash.Netlist.Id as Id
 import           Clash.Netlist.Util (typeSize)
@@ -66,7 +66,9 @@ import           Clash.Primitives.Util (hashCompiledPrimMap)
 import           Clash.Signal (VDomainConfiguration(..))
 import           Clash.Util.Graph (callGraphBindings)
 
-#if MIN_VERSION_ghc(9,0,0)
+#if MIN_VERSION_ghc(9,4,0)
+import GHC.Data.Bool (OverridingBool(..))
+#elif MIN_VERSION_ghc(9,0,0)
 import GHC.Utils.Misc (OverridingBool(..))
 #else
 import Util (OverridingBool(..))
@@ -164,10 +166,8 @@ data Manifest
     -- ^ Dependencies of this design (fully qualified binder names). Is a
     -- transitive closure of all dependencies.
     --
-    -- This list is topologically sorted. I.e., a dependency might depend
-    -- on any dependency listed after it, but not before it.
-    --
-    -- TODO: this ordered differs from `fileNames` and `componentNames`. Fix?
+    -- This list is reverse topologically sorted. I.e., a component might depend
+    -- on any component listed before it, but not after it.
   } deriving (Show,Read,Eq)
 
 instance ToJSON Manifest where
@@ -291,7 +291,7 @@ mkManifestPort backend portId portType portDir = ManifestPort{..}
   mpName = Id.toText portId
   mpWidth = typeSize portType
   mpDirection = portDir
-  mpIsClock = case portType of {Clock _ -> True; _ -> False}
+  mpIsClock = case portType of {Clock _ -> True; ClockN _ -> True; _ -> False}
   mpDomain = hwTypeDomain portType
   mpTypeName = flip evalState backend $ getAp $ do
      LText.toStrict . renderOneLine <$> hdlType (External mpName) portType
