@@ -1,8 +1,9 @@
 {-|
   Copyright  :  (C) 2013-2016, University of Twente,
                     2016-2017, Myrtle Software Ltd,
-                    2017     , QBayLogic, Google Inc.
-                    2020-2022, QBayLogic
+                    2017     , QBayLogic, Google Inc.,
+                    2020-2022, QBayLogic,
+                    2022     , Google Inc.,
   License    :  BSD2 (see the file LICENSE)
   Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -41,7 +42,11 @@ import           Data.Text.Prettyprint.Doc
 
 import           GHC.Generics                   (Generic)
 
-#if MIN_VERSION_ghc(9,0,0)
+#if MIN_VERSION_ghc(9,4,0)
+import           GHC.Types.Basic                (InlineSpec)
+import           GHC.Types.SrcLoc               (SrcSpan)
+import           GHC.Data.Bool                  (OverridingBool(..))
+#elif MIN_VERSION_ghc(9,0,0)
 import           GHC.Types.Basic                (InlineSpec)
 import           GHC.Types.SrcLoc               (SrcSpan)
 import           GHC.Utils.Misc                 (OverridingBool(..))
@@ -54,6 +59,7 @@ import           Util                           (OverridingBool(..))
 import           Clash.Annotations.BitRepresentation.Internal (CustomReprs)
 import           Clash.Signal.Internal
 
+import           Clash.Backend.Verilog.Time     (Period(..), Unit(Fs))
 import           Clash.Core.Term                (Term)
 import           Clash.Core.TyCon               (TyConMap, TyConName)
 import           Clash.Core.Var                 (Id)
@@ -68,18 +74,17 @@ data ClashEnv = ClashEnv
   , envTupleTyCons :: IntMap TyConName
   , envPrimitives  :: CompiledPrimMap
   , envCustomReprs :: CustomReprs
+  , envDomains     :: DomainMap
   } deriving (Generic, NFData)
 
 data ClashDesign = ClashDesign
   { designEntities :: [TopEntityT]
-  , designDomains  :: DomainMap
   , designBindings :: BindingMap
   }
 
 instance NFData ClashDesign where
   rnf design =
     designEntities design `seq`
-    designDomains design `deepseq`
     designBindings design `deepseq`
     ()
 
@@ -340,7 +345,7 @@ data ClashOpts = ClashOpts
   --
   -- Command line flag: -fdiagnostics-color
   , opt_intWidth :: Int
-  -- ^ Set the bit width for the Int/Word/Integer types. The only allowed values
+  -- ^ Set the bit width for the Int\/Word\/Integer types. The only allowed values
   -- are 32 or 64.
   , opt_hdlDir :: Maybe String
   -- ^ Directory to save HDL files to
@@ -358,8 +363,8 @@ data ClashOpts = ClashOpts
   , opt_escapedIds :: Bool
   -- ^ Use escaped identifiers in HDL. See:
   --
-  --  * http://vhdl.renerta.com/mobile/source/vhd00037.htm
-  --  * http://verilog.renerta.com/source/vrg00018.htm
+  --  * https://peterfab.com/ref/vhdl/vhdl_renerta/source/vhd00037.htm
+  --  * https://peterfab.com/ref/verilog/verilog_renerta/source/vrg00018.htm
   , opt_lowerCaseBasicIds :: PreserveCase
   -- ^ Force all generated basic identifiers to lowercase. Among others, this
   -- affects module and file names.
@@ -392,6 +397,9 @@ data ClashOpts = ClashOpts
   , opt_renderEnums :: Bool
   -- ^ Render sum types with all zero-width fields as enums where supported, as
   -- opposed to rendering them as bitvectors.
+  , opt_timescalePrecision :: Period
+  -- ^ Timescale precision set in Verilog files. E.g., setting this would sets
+  -- the second part of @`timescale 100fs/100fs@.
   }
   deriving (Show)
 
@@ -424,6 +432,7 @@ instance NFData ClashOpts where
     opt_inlineWFCacheLimit o `deepseq`
     opt_edalize o `deepseq`
     opt_renderEnums o `deepseq`
+    opt_timescalePrecision o `deepseq`
     ()
 
 instance Eq ClashOpts where
@@ -454,7 +463,8 @@ instance Eq ClashOpts where
     opt_aggressiveXOptBB s0 == opt_aggressiveXOptBB s1 &&
     opt_inlineWFCacheLimit s0 == opt_inlineWFCacheLimit s1 &&
     opt_edalize s0 == opt_edalize s1 &&
-    opt_renderEnums s0 == opt_renderEnums s1
+    opt_renderEnums s0 == opt_renderEnums s1 &&
+    opt_timescalePrecision s0 == opt_timescalePrecision s1
 
    where
     eqOverridingBool :: OverridingBool -> OverridingBool -> Bool
@@ -492,7 +502,8 @@ instance Hashable ClashOpts where
     opt_aggressiveXOptBB `hashWithSalt`
     opt_inlineWFCacheLimit `hashWithSalt`
     opt_edalize `hashWithSalt`
-    opt_renderEnums
+    opt_renderEnums `hashWithSalt`
+    opt_timescalePrecision
    where
     hashOverridingBool :: Int -> OverridingBool -> Int
     hashOverridingBool s1 Auto = hashWithSalt s1 (0 :: Int)
@@ -531,6 +542,7 @@ defClashOpts
   , opt_inlineWFCacheLimit  = 10 -- TODO: find "optimal" value
   , opt_edalize             = False
   , opt_renderEnums         = True
+  , opt_timescalePrecision  = Period 100 Fs
   }
 
 -- | Synopsys Design Constraint (SDC) information for a component.
