@@ -3,6 +3,7 @@
                      2016-2017, Myrtle Software Ltd,
                      2017-2018, Google Inc.,
                      2021-2022, QBayLogic B.V.
+                     2022     , Google Inc.
   License     :  BSD2 (see the file LICENSE)
   Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -24,18 +25,20 @@ module Clash.Netlist where
 import           Control.Exception                (throw)
 import           Control.Lens                     ((.=), (<~))
 import qualified Control.Lens                     as Lens
-import           Control.Monad.Extra              (concatMapM)
+import           Control.Monad                    (zipWithM)
+import           Control.Monad.Extra              (concatMapM, mapMaybeM)
 import           Control.Monad.Reader             (runReaderT)
 import           Control.Monad.State.Strict       (State, runStateT, runState)
 import           Data.Bifunctor                   (first, second)
 import           Data.Char                        (ord)
 import           Data.Either                      (partitionEithers, rights)
 import           Data.Foldable                    (foldlM)
-import qualified Data.HashMap.Lazy                as HashMap
-import           Data.List                        (elemIndex, partition, sortOn)
+import           Data.List                        (elemIndex, partition)
 import           Data.List.Extra                  (zipEqual)
+import           Data.List.NonEmpty               (NonEmpty (..))
+import qualified Data.List.NonEmpty.Extra         as NE
 import           Data.Maybe
-  (listToMaybe, mapMaybe, fromMaybe)
+  (listToMaybe, fromMaybe)
 import qualified Data.Map.Ordered                 as OMap
 import qualified Data.Set                         as Set
 import           Data.Primitive.ByteArray         (ByteArray (..))
@@ -55,7 +58,7 @@ import           Outputable                       (ppr, showSDocUnsafe)
 import           SrcLoc                           (isGoodSrcSpan)
 #endif
 
-import           Clash.Annotations.Primitive      (extractPrim, HDL)
+import           Clash.Annotations.Primitive      (HDL)
 import           Clash.Annotations.BitRepresentation.ClashLib
   (coreToType')
 import           Clash.Annotations.BitRepresentation.Internal
@@ -172,6 +175,7 @@ runNetlistMonad env isTb s tops typeTrans ite be seenIds_ dir componentNames_
         , _backEndITE=ite
         , _backend=be
         , _htyCache=mempty
+        , _usages=mempty
         }
 
 -- | Generate names for all binders in "BindingMap", except for the ones already
@@ -255,6 +259,7 @@ genComponentT compName0 componentExpr = do
   compName1 <- (`lookupVarEnv'` compName0) <$> Lens.use componentNames
   sp <- (bindingLoc . (`lookupVarEnv'` compName0)) <$> Lens.use bindings
   curCompNm .= (compName1, sp)
+  usages .= mempty
 
   topEntityTM <- lookupVarEnv compName0 <$> Lens.use topEntityAnns
   let topAnnMM = topAnnotation <$> topEntityTM
@@ -280,29 +285,33 @@ genComponentT compName0 componentExpr = do
 
   case resultM of
     Just result -> do
-      [NetDecl' _ rw _ _ rIM] <- case filter ((==result) . fst) binders of
+      [NetDecl' _ _ _ rIM] <- case filter ((==result) . fst) binders of
         b:_ -> mkNetDecl b
         _ -> error "internal error: couldn't find result binder"
 
+      u <- Lens.use usages
+      let useOf i = fromMaybe Cont $ lookupUsage (fst i) u
+
       let (compOutps',resUnwrappers') = case compOutps of
-            [oport] -> ([(rw,oport,rIM)],resUnwrappers)
-            _       -> let NetDecl n res resTy = case resUnwrappers of
-                             decl:_ -> decl
-                             _ -> error "internal error: insufficient resUnwrappers"
-                       in  (map (Wire,,Nothing) compOutps
-                           ,NetDecl' n rw res (Right resTy) Nothing:tail resUnwrappers
-                           )
+            [oport] -> ([(useOf oport,oport,rIM)],resUnwrappers)
+            _ -> case resUnwrappers of
+              NetDecl n res resTy:_ ->
+                (map (\op -> (useOf op,op,Nothing)) compOutps
+                ,NetDecl' n res resTy Nothing:drop 1 resUnwrappers
+                )
+              _ -> error "internal error: insufficient resUnwrappers"
           component      = Component compName1 compInps compOutps'
                              (netDecls ++ argWrappers ++ decls ++ resUnwrappers')
       ids <- Lens.use seenIds
-      return (ComponentMeta wereVoids sp ids, component)
+      return (ComponentMeta wereVoids sp ids u, component)
     -- No result declaration means that the result is empty, this only happens
     -- when the TopEntity has an empty result. We just create an empty component
     -- in this case.
     Nothing -> do
       let component = Component compName1 compInps [] (netDecls ++ argWrappers ++ decls)
       ids <- Lens.use seenIds
-      return (ComponentMeta wereVoids sp ids, component)
+      u <- Lens.use usages
+      return (ComponentMeta wereVoids sp ids u, component)
 
 mkNetDecl :: (Id, Term) -> NetlistMonad [Declaration]
 mkNetDecl (id_,tm) = preserveVarEnv $ do
@@ -321,19 +330,16 @@ mkNetDecl (id_,tm) = preserveVarEnv $ do
         resInits1 = map Just resInits0 <> repeat Nothing
         mpInfo = multiPrimInfo' tcm pInfo
         (_, res) = splitMultiPrimArgs mpInfo args0
+
         netdecl i typ resInit =
-          -- TODO: Dehardcode Wire. Would entail changing 'outputReg' to a
-          -- list.
-          NetDecl' srcNote Wire (id2identifier i) (Right typ) resInit
+          NetDecl' srcNote (Id.unsafeFromCoreId i) typ resInit
 
       hwTys <- mapM (unsafeCoreTypeToHWTypeM' $(curLoc)) (mpi_resultTypes mpInfo)
       pure (zipWith3 netdecl res hwTys resInits1)
 
-
     singleDecl hwTy = do
-      wr <- termToWireOrReg tm
       rIM <- listToMaybe <$> getResInits (id_, tm)
-      return (NetDecl' srcNote wr (id2identifier id_) (Right hwTy) rIM)
+      return (NetDecl' srcNote (Id.unsafeFromCoreId id_) hwTy rIM)
 
     addSrcNote loc
       | isGoodSrcSpan loc = Just (StrictText.pack (showSDocUnsafe (ppr loc)))
@@ -354,23 +360,6 @@ mkNetDecl (id_,tm) = preserveVarEnv $ do
       | isMultiPrimSelect t = False
       | otherwise = True
 
-    termToWireOrReg :: Term -> NetlistMonad WireOrReg
-    termToWireOrReg (stripTicks -> Case scrut _ alts0@(_:_:_)) = do
-      tcm <- Lens.view tcCache
-      let scrutTy = inferCoreTypeOf tcm scrut
-      scrutHTy <- unsafeCoreTypeToHWTypeM' $(curLoc) scrutTy
-      ite <- Lens.use backEndITE
-      case iteAlts scrutHTy alts0 of
-        Just _ | ite -> return Wire
-        _ -> return Reg
-    termToWireOrReg (collectArgs -> (Prim p,_)) = do
-      bbM <- HashMap.lookup (primName p) <$> Lens.view primitives
-      case bbM of
-        Just (extractPrim -> Just BlackBox {..}) | outputReg -> return Reg
-        _ | primName p == "Clash.Explicit.SimIO.mealyIO" -> return Reg
-        _ -> return Wire
-    termToWireOrReg _ = return Wire
-
     -- Set the initialization value of a signal when a primitive wants to set it
     getResInits :: (Id, Term) -> NetlistMonad [Expr]
     getResInits (i,collectArgsTicks -> (k,args0,ticks)) = case k of
@@ -431,8 +420,7 @@ mkDeclarations'
   -- ^ RHS of the let-binder
   -> NetlistMonad [Declaration]
 mkDeclarations' _declType bndr (collectTicks -> (Var v,ticks)) =
-  withTicks ticks $ \tickDecls -> do
-  mkFunApp (id2identifier bndr) v [] tickDecls
+  withTicks ticks (mkFunApp (Id.unsafeFromCoreId bndr) v [])
 
 mkDeclarations' _declType _bndr e@(collectTicks -> (Case _ _ [],_)) = do
   (_,sp) <- Lens.use curCompNm
@@ -445,37 +433,42 @@ mkDeclarations' _declType _bndr e@(collectTicks -> (Case _ _ [],_)) = do
                     ])
           Nothing
 
-mkDeclarations' declType bndr (collectTicks -> (Case scrut altTy alts@(_:_:_),ticks)) =
-  withTicks ticks $ \tickDecls -> do
-  mkSelection declType (CoreId bndr) scrut altTy alts tickDecls
+mkDeclarations' declType bndr (collectTicks -> (Case scrut altTy (alt:alts@(_:_)),ticks)) =
+  withTicks ticks (mkSelection declType (CoreId bndr) scrut altTy (alt :| alts))
 
 mkDeclarations' declType bndr app = do
   let (appF,args0,ticks) = collectArgsTicks app
       (args,tyArgs) = partitionEithers args0
   case appF of
     Var f
-      | null tyArgs -> withTicks ticks (mkFunApp (id2identifier bndr) f args)
+      | null tyArgs ->
+        withTicks ticks (mkFunApp (Id.unsafeFromCoreId bndr) f args)
       | otherwise   -> do
         (_,sp) <- Lens.use curCompNm
         throw (ClashException sp ($(curLoc) ++ "Not in normal form: Var-application with Type arguments:\n\n" ++ showPpr app) Nothing)
     _ -> do
       (exprApp,declsApp0) <- mkExpr False declType (CoreId bndr) app
-      let dstId = id2identifier bndr
-          assn  =
-            case exprApp of
-              Identifier _ Nothing ->
-                -- Supplied 'bndr' was used to assign a result to, so we
-                -- don't have to manually turn it into a declaration
-                []
-              Noop ->
-                -- Rendered expression rendered a "noop" - a list of
-                -- declarations without a result. Used for things like
-                -- mealy IO / inline assertions / multi result primitives.
-                []
-              _ ->
-                -- Turn returned expression into declaration by assigning
-                -- it to 'dstId'
-                [Assignment dstId exprApp]
+      let dstId = Id.unsafeFromCoreId bndr
+      assn  <- case exprApp of
+                 Identifier _ Nothing ->
+                   -- Supplied 'bndr' was used to assign a result to, so we
+                   -- don't have to manually turn it into a declaration
+                   pure []
+
+                 Noop ->
+                   -- Rendered expression rendered a "noop" - a list of
+                   -- declarations without a result. Used for things like
+                   -- mealy IO / inline assertions / multi result primitives.
+                   pure []
+
+                 _ -> do
+                   -- Turn returned expression into declaration by assigning
+                   -- it to 'dstId'
+                   assn <- case declType of
+                     Concurrent -> contAssign dstId exprApp
+                     Sequential -> procAssign Blocking dstId exprApp
+                   pure [assn]
+
       declsApp1 <- if null declsApp0
                    then withTicks ticks return
                    else pure declsApp0
@@ -488,11 +481,11 @@ mkSelection
   -> NetlistId
   -> Term
   -> Type
-  -> [Alt]
+  -> NonEmpty Alt
   -> [Declaration]
   -> NetlistMonad [Declaration]
 mkSelection declType bndr scrut altTy alts0 tickDecls = do
-  let dstId = netlistId1 id id2identifier bndr
+  let dstId = netlistId1 id Id.unsafeFromCoreId bndr
   tcm <- Lens.view tcCache
   let scrutTy = inferCoreTypeOf tcm scrut
   scrutHTy <- unsafeCoreTypeToHWTypeM' $(curLoc) scrutTy
@@ -500,13 +493,13 @@ mkSelection declType bndr scrut altTy alts0 tickDecls = do
   (_,sp) <- Lens.use curCompNm
   ite <- Lens.use backEndITE
   altHTy <- unsafeCoreTypeToHWTypeM' $(curLoc) altTy
-  case iteAlts scrutHTy alts0 of
+  case iteAlts scrutHTy (NE.toList alts0) of
     Just (altT,altF)
       | ite
       , Concurrent <- declType
       -> do
       (scrutExpr,scrutDecls) <- case scrutHTy of
-        SP {} -> first (mkScrutExpr sp scrutHTy (fst (last alts0))) <$>
+        SP {} -> first (mkScrutExpr sp scrutHTy (fst (NE.last alts0))) <$>
                    mkExpr True declType (NetlistId scrutId scrutTy) scrut
         _ -> mkExpr False declType (NetlistId scrutId scrutTy) scrut
       altTId <- Id.suffix dstId "sel_alt_t"
@@ -525,19 +518,18 @@ mkSelection declType bndr scrut altTy alts0 tickDecls = do
          | isVoid altHTy
           -> return $! altTDecls ++ altFDecls
          | otherwise
-          -> return $! scrutDecls ++ altTDecls ++ altFDecls ++ tickDecls ++
-                [Assignment dstId (IfThenElse scrutExpr altTExpr altFExpr)]
+          -> do dstAssign <- contAssign dstId (IfThenElse scrutExpr altTExpr altFExpr)
+                return $! scrutDecls ++ altTDecls ++ altFDecls ++ tickDecls ++ [dstAssign]
     _ -> do
       reprs <- Lens.view customReprs
       let alts1 = (reorderDefault . reorderCustom tcm reprs scrutTy) alts0
-      (scrutExpr,scrutDecls) <- first (mkScrutExpr sp scrutHTy (fst (head alts1))) <$>
+      (scrutExpr,scrutDecls) <- first (mkScrutExpr sp scrutHTy (fst (NE.head alts1))) <$>
                                   mkExpr True declType (NetlistId scrutId scrutTy) scrut
-      (exprs,altsDecls)      <- unzip <$> mapM (mkCondExpr scrutHTy) alts1
+      (exprs,altsDecls)      <- unzip <$> mapM (mkCondExpr scrutHTy) (NE.toList alts1)
       case declType of
         Sequential -> do
           -- Assign to the result in every branch
-          let (altNets,exprAlts) = unzip (zipWith (altAssign dstId)
-                                                  exprs altsDecls)
+          (altNets,exprAlts) <- fmap unzip (zipWithM (altAssign dstId) exprs altsDecls)
           return $! scrutDecls ++ tickDecls ++ concat altNets ++
                     [Seq [Branch scrutExpr scrutHTy exprAlts]]
         Concurrent ->
@@ -546,12 +538,12 @@ mkSelection declType bndr scrut altTy alts0 tickDecls = do
              | isVoid altHTy
               -> return $! concat altsDecls
              | otherwise
-              -> return $! scrutDecls ++ concat altsDecls ++ tickDecls
-                    ++ [CondAssignment dstId altHTy scrutExpr scrutHTy exprs]
+              -> do assign <- condAssign dstId altHTy scrutExpr scrutHTy exprs
+                    return $! scrutDecls ++ concat altsDecls ++ tickDecls ++ [assign]
  where
   mkCondExpr :: HWType -> (Pat,Term) -> NetlistMonad ((Maybe HW.Literal,Expr),[Declaration])
   mkCondExpr scrutHTy (pat,alt) = do
-    altId <- Id.suffix (netlistId1 id id2identifier bndr) "sel_alt"
+    altId <- Id.suffix (netlistId1 id Id.unsafeFromCoreId bndr) "sel_alt"
     (altExpr,altDecls) <- mkExpr False declType (NetlistId altId altTy) alt
     (,altDecls) <$> case pat of
       DefaultPat           -> return (Nothing,altExpr)
@@ -562,6 +554,14 @@ mkSelection declType bndr scrut altTy alts0 tickDecls = do
       LitPat  (CharLiteral c) -> return (Just (NumLit . toInteger $ ord c), altExpr)
       LitPat  (Int64Literal i) -> return (Just (NumLit i), altExpr)
       LitPat  (Word64Literal w) -> return (Just (NumLit w), altExpr)
+#if MIN_VERSION_base(4,16,0)
+      LitPat  (Int8Literal i) -> return (Just (NumLit i), altExpr)
+      LitPat  (Int16Literal i) -> return (Just (NumLit i), altExpr)
+      LitPat  (Int32Literal i) -> return (Just (NumLit i), altExpr)
+      LitPat  (Word8Literal w) -> return (Just (NumLit w), altExpr)
+      LitPat  (Word16Literal w) -> return (Just (NumLit w), altExpr)
+      LitPat  (Word32Literal w) -> return (Just (NumLit w), altExpr)
+#endif
       LitPat  (NaturalLiteral n) -> return (Just (NumLit n), altExpr)
       _  -> do
         (_,sp) <- Lens.use curCompNm
@@ -575,12 +575,18 @@ mkSelection declType bndr scrut altTy alts0 tickDecls = do
                           _ -> throw (ClashException sp ($(curLoc) ++ "Not in normal form: Not a variable reference or primitive as subject of a case-statement:\n\n" ++ show scrutE) Nothing)
     _ -> scrutE
 
-  altAssign :: Identifier -> (Maybe HW.Literal,Expr) -> [Declaration]
-            -> ([Declaration],(Maybe HW.Literal,[Seq]))
-  altAssign i (m,expr) ds =
+  altAssign
+    :: Identifier
+    -> (Maybe HW.Literal,Expr)
+    -> [Declaration]
+    -> NetlistMonad ([Declaration],(Maybe HW.Literal,[Seq]))
+  altAssign i (m,expr) ds = do
     let (nets,rest) = partition isNet ds
-        assn = case expr of { Noop -> []; _ -> [SeqDecl (Assignment i expr)] }
-    in  (nets,(m,map SeqDecl rest ++ assn))
+    assn <- case expr of
+              Noop -> pure []
+              _ -> do assn <- procAssign Blocking i expr
+                      pure [assn]
+    pure (nets,(m,map SeqDecl (rest ++ assn)))
    where
     isNet NetDecl' {} = True
     isNet _ = False
@@ -588,23 +594,26 @@ mkSelection declType bndr scrut altTy alts0 tickDecls = do
 -- GHC puts default patterns in the first position, we want them in the
 -- last position.
 reorderDefault
-  :: [(Pat, Term)]
-  -> [(Pat, Term)]
-reorderDefault ((DefaultPat,e):alts') = alts' ++ [(DefaultPat,e)]
-reorderDefault alts'                  = alts'
+  :: NonEmpty (Pat, Term)
+  -> NonEmpty (Pat, Term)
+reorderDefault ((DefaultPat,e) :| alts') =
+  case alts' of
+    [] -> (DefaultPat,e) :| []
+    x:xs -> x :| (xs <> [(DefaultPat,e)])
+reorderDefault alts' = alts'
 
 reorderCustom
   :: TyConMap
   -> CustomReprs
   -> Type
-  -> [(Pat, Term)]
-  -> [(Pat, Term)]
+  -> NonEmpty (Pat, Term)
+  -> NonEmpty (Pat, Term)
 reorderCustom tcm reprs (coreView1 tcm -> Just ty) alts =
   reorderCustom tcm reprs ty alts
 reorderCustom _tcm reprs (coreToType' -> Right typeName) alts =
   case getDataRepr typeName reprs of
     Just (DataRepr' _name _size _constrReprs) ->
-      sortOn (patPos reprs . fst) alts
+      NE.sortOn (patPos reprs . fst) alts
     Nothing ->
       alts
 reorderCustom _tcm _reprs _type alts =
@@ -691,14 +700,13 @@ mkFunApp dstId fun args tickDecls = do
         --       pure (expandTopEntityOrErr is (zip argIds argTys) (resId, resTy) topAnnM)
 
         -- Generate ExpandedTopEntity, see TODO^
-        is <- Lens.use seenIds
         argTys <- mapM (unsafeCoreTypeToHWTypeM $(curLoc) . inferCoreTypeOf tcm) args
         resTy <- unsafeCoreTypeToHWTypeM $(curLoc) fResTy
         let
           ettArgs = (Nothing,) <$> argTys
           ettRes = (Nothing, resTy)
-          expandedTopEntity =
-            expandTopEntityOrErr is ettArgs ettRes (topAnnotation topEntity)
+        expandedTopEntity <-
+            expandTopEntityOrErrM ettArgs ettRes (topAnnotation topEntity)
 
         instDecls <-
           mkTopUnWrapper
@@ -744,6 +752,7 @@ mkFunApp dstId fun args tickDecls = do
               instLabel3 <- Id.makeBasic instLabel2
               let portMap = NamedPortMap (outpAssign ++ inpAssigns)
                   instDecl = InstDecl Entity Nothing [] compName instLabel3 [] portMap
+              declareInstUses outpAssign
               return (argDecls ++ argDecls' ++ tickDecls ++ [instDecl])
             else
               let
@@ -769,9 +778,10 @@ mkFunApp dstId fun args tickDecls = do
             |]
     _ ->
       case args of
-        [] ->
+        [] -> do
           -- TODO: Figure out what to do with zero-width constructs
-          return [Assignment dstId (Identifier (id2identifier fun) Nothing)]
+          assn <- contAssign dstId (Identifier (Id.unsafeFromCoreId fun) Nothing)
+          pure [assn]
         _ -> error [I.i|
           Netlist generation encountered a local function. This should not
           happen. Function:
@@ -802,9 +812,8 @@ toSimpleVar _ (e@(Identifier _ Nothing),_) = return (e,[])
 toSimpleVar dstId (e,ty) = do
   argNm <- Id.suffix dstId "fun_arg"
   hTy <- unsafeCoreTypeToHWTypeM' $(curLoc) ty
-  let argDecl         = NetDecl Nothing argNm hTy
-      argAssn         = Assignment argNm e
-  return (Identifier argNm Nothing,[argDecl,argAssn])
+  argDecl <- mkInit Concurrent Cont argNm hTy e
+  return (Identifier argNm Nothing, argDecl)
 
 -- | Generate an expression for a term occurring on the RHS of a let-binder
 mkExpr :: HasCallStack
@@ -822,6 +831,14 @@ mkExpr _ _ _ (stripTicks -> Core.Literal l) = do
     WordLiteral w    -> return (HW.Literal (Just (Unsigned iw,iw)) $ NumLit w, [])
     Int64Literal i   -> return (HW.Literal (Just (Signed 64,64)) $ NumLit i, [])
     Word64Literal w  -> return (HW.Literal (Just (Unsigned 64,64)) $ NumLit w, [])
+#if MIN_VERSION_ghc(8,8,0)
+    Int8Literal i    -> return (HW.Literal (Just (Signed 8,8)) $ NumLit i, [])
+    Int16Literal i   -> return (HW.Literal (Just (Signed 16,16)) $ NumLit i, [])
+    Int32Literal i   -> return (HW.Literal (Just (Signed 32,32)) $ NumLit i, [])
+    Word8Literal w   -> return (HW.Literal (Just (Unsigned 8,8)) $ NumLit w, [])
+    Word16Literal w  -> return (HW.Literal (Just (Unsigned 16,16)) $ NumLit w, [])
+    Word32Literal w  -> return (HW.Literal (Just (Unsigned 32,32)) $ NumLit w, [])
+#endif
     CharLiteral c    -> return (HW.Literal (Just (Unsigned 21,21)) . NumLit . toInteger $ ord c, [])
     FloatLiteral w   -> return (HW.Literal (Just (BitVector 32,32)) (NumLit $ toInteger w), [])
     DoubleLiteral w  -> return (HW.Literal (Just (BitVector 64,64)) (NumLit $ toInteger w), [])
@@ -843,45 +860,40 @@ mkExpr bbEasD declType bndr app =
         hwTy:_ -> hwTy
         _ -> error ("internal error: unable to extract sufficient hwTys from: " <> show bndr)
   case appF of
-    Data dc -> mkDcApplication hwTys bndr dc tmArgs
-    Prim pInfo -> mkPrimitive False bbEasD bndr pInfo args tickDecls
+    Data dc -> mkDcApplication declType hwTys bndr dc tmArgs
+    Prim pInfo -> mkPrimitive False bbEasD declType bndr pInfo args tickDecls
     Var f
       | null tmArgs ->
           if isVoid hwTyA then
             return (Noop, [])
-          else
-            return (Identifier (id2identifier f) Nothing, [])
+          else do
+            return (Identifier (Id.unsafeFromCoreId f) Nothing, [])
       | not (null tyArgs) ->
           throw (ClashException sp ($(curLoc) ++ "Not in normal form: "
             ++ "Var-application with Type arguments:\n\n" ++ showPpr app) Nothing)
       | otherwise -> do
-          argNm <- Id.suffix (netlistId1 id id2identifier bndr) "fun_arg"
+          argNm <- Id.suffix (netlistId1 id Id.unsafeFromCoreId bndr) "fun_arg"
           decls  <- mkFunApp argNm f tmArgs tickDecls
           if isVoid hwTyA then
             return (Noop, decls)
           else
+            -- This net was already declared in the call to mkSelection.
             return ( Identifier argNm Nothing
-                   , NetDecl' Nothing Wire argNm (Right hwTyA) Nothing:decls)
+                   , NetDecl Nothing argNm hwTyA : decls)
     Case scrut ty' [alt] -> mkProjection bbEasD bndr scrut ty' alt
-    Case scrut tyA alts -> do
-      tcm <- Lens.view tcCache
-      let scrutTy = inferCoreTypeOf tcm scrut
-      scrutHTy <- unsafeCoreTypeToHWTypeM' $(curLoc) scrutTy
-      ite <- Lens.use backEndITE
-      let wr = case iteAlts scrutHTy alts of
-                 Just _ | ite -> Wire
-                 _ -> Reg
-      argNm <- Id.suffix (netlistId1 id id2identifier bndr) "sel_arg"
+    Case scrut tyA (alt:alts) -> do
+      argNm <- Id.suffix (netlistId1 id Id.unsafeFromCoreId bndr) "sel_arg"
       decls  <- mkSelection declType (NetlistId argNm (netlistTypes1 bndr))
-                            scrut tyA alts tickDecls
+                            scrut tyA (alt :| alts) tickDecls
       if isVoid hwTyA then
         return (Noop, decls)
       else
+        -- This net was already declared in the call to mkSelection
         return ( Identifier argNm Nothing
-               , NetDecl' Nothing wr argNm (Right hwTyA) Nothing:decls)
+               , NetDecl' Nothing argNm hwTyA Nothing:decls)
     Letrec binders body -> do
       netDecls <- concatMapM mkNetDecl binders
-      decls    <- concatMapM (uncurry mkDeclarations) binders
+      decls    <- concatMapM (uncurry (mkDeclarations' declType)) binders
       (bodyE,bodyDecls) <- mkExpr bbEasD declType bndr (mkApps (mkTicks body ticks) args)
       return (bodyE,netDecls ++ decls ++ bodyDecls)
     _ -> throw (ClashException sp ($(curLoc) ++ "Not in normal form: application of a Lambda-expression\n\n" ++ showPpr app) Nothing)
@@ -917,7 +929,7 @@ mkProjection mkDec bndr scrut altTy alt@(pat,v) = do
     scrutNm <-
       netlistId1
         Id.next
-        (\b -> Id.suffix (id2identifier b) "projection")
+        (\b -> Id.suffix (Id.unsafeFromCoreId b) "projection")
         bndr
     (scrutExpr,newDecls) <- mkExpr False Concurrent (NetlistId scrutNm scrutTy) scrut
     case scrutExpr of
@@ -930,14 +942,13 @@ mkProjection mkDec bndr scrut altTy alt@(pat,v) = do
         -- TODO: seems useless?
         pure (Left newDecls)
       _ -> do
-        let scrutDecl = NetDecl Nothing scrutNm sHwTy
-            scrutAssn = Assignment scrutNm scrutExpr
-        pure (Right (scrutNm, Nothing, newDecls ++ [scrutDecl, scrutAssn]))
+        scrutDecl <- mkInit Concurrent Cont scrutNm sHwTy scrutExpr
+        pure (Right (scrutNm, Nothing, newDecls ++ scrutDecl))
 
   case scrutRendered of
     Left newDecls -> pure (Noop, newDecls)
     Right (selId, modM, decls) -> do
-      let altVarId = id2identifier varTm
+      let altVarId = Id.unsafeFromCoreId varTm
       modifier <- case pat of
         DataPat dc exts tms -> do
           let
@@ -968,9 +979,8 @@ mkProjection mkDec bndr scrut altTy alt@(pat,v) = do
       case bndr of
         NetlistId scrutNm _ | mkDec -> do
           scrutNm' <- Id.next scrutNm
-          let scrutDecl = NetDecl Nothing scrutNm' vHwTy
-              scrutAssn = Assignment scrutNm' extractExpr
-          return (Identifier scrutNm' Nothing,scrutDecl:scrutAssn:decls)
+          scrutDecl <- mkInit Concurrent Cont scrutNm' vHwTy extractExpr
+          return (Identifier scrutNm' Nothing, scrutDecl ++ decls)
         MultiId {} -> error "mkProjection: MultiId"
         _ -> return (extractExpr,decls)
   where
@@ -981,7 +991,8 @@ mkProjection mkDec bndr scrut altTy alt@(pat,v) = do
 -- | Generate an expression for a DataCon application occurring on the RHS of a let-binder
 mkDcApplication
     :: HasCallStack
-    => [HWType]
+    => DeclarationType
+    -> [HWType]
     -- ^ HWType of the LHS of the let-binder, can multiple types when we're
     -- creating a "split" product type (e.g. a tuple of a Clock and Reset)
     -> NetlistId
@@ -992,15 +1003,15 @@ mkDcApplication
     -- ^ DataCon Arguments
     -> NetlistMonad (Expr,[Declaration])
     -- ^ Returned expression and a list of generate BlackBox declarations
-mkDcApplication [dstHType] bndr dc args = do
+mkDcApplication declType [dstHType] bndr dc args = do
   let dcNm = nameOcc (dcName dc)
   tcm <- Lens.view tcCache
   let argTys = map (inferCoreTypeOf tcm) args
-  argNm <- netlistId1 return (\b -> Id.suffix (id2identifier b) "_dc_arg") bndr
+  argNm <- netlistId1 return (\b -> Id.suffix (Id.unsafeFromCoreId b) "_dc_arg") bndr
   argHWTys <- mapM coreTypeToHWTypeM' argTys
 
   (argExprs, concat -> argDecls) <- unzip <$>
-    mapM (\(e,t) -> mkExpr False Concurrent (NetlistId argNm t) e) (zip args argTys)
+    mapM (\(e,t) -> mkExpr False declType (NetlistId argNm t) e) (zip args argTys)
 
   -- Filter void arguments, but make sure to render their declarations:
   let
@@ -1127,7 +1138,7 @@ mkDcApplication [dstHType] bndr dc args = do
         error $ $(curLoc) ++ "mkDcApplication undefined for: " ++ show (dstHType,dc,args,argHWTys)
 
 -- Handle MultiId assignment
-mkDcApplication dstHTypes (MultiId argNms) _ args = do
+mkDcApplication declType dstHTypes (MultiId argNms) _ args = do
   tcm                 <- Lens.view tcCache
   let argTys          = map (inferCoreTypeOf tcm) args
   argHWTys            <- mapM coreTypeToHWTypeM' argTys
@@ -1137,18 +1148,20 @@ mkDcApplication dstHTypes (MultiId argNms) _ args = do
       (_hWTysFiltered,argsFiltered) = unzip
         (filter (maybe True (not . isVoid) . fst) argsBundled)
   (argExprs,argDecls) <- fmap (second concat . unzip) $!
-                         mapM (uncurry (mkExpr False Concurrent)) argsFiltered
+                         mapM (uncurry (mkExpr False declType)) argsFiltered
   if length dstHTypes == length argExprs then do
-    let assns = mapMaybe
-                  (\case (_,Noop) -> Nothing
-                         (dstId,e) -> let nm = netlistId1 id id2identifier dstId
-                                      in  case e of
+    assns <- mapMaybeM
+                  (\case (_,Noop) -> pure Nothing
+                         (dstId,e) -> let nm = netlistId1 id Id.unsafeFromCoreId dstId
+                                       in case e of
                                             Identifier nm0 Nothing
-                                              | nm == nm0 -> Nothing
-                                            _ -> Just (Assignment nm e))
+                                              | nm == nm0 -> pure Nothing
+                                            _ -> Just <$> case declType of
+                                                            Concurrent -> contAssign nm e
+                                                            Sequential -> procAssign Blocking nm e)
                   (zipEqual (map CoreId argNms) argExprs)
     return (Noop,argDecls ++ assns)
   else
     error "internal error"
 
-mkDcApplication _ _ _ _ = error "internal error"
+mkDcApplication _ _ _ _ _ = error "internal error"
