@@ -57,7 +57,9 @@ import Data.Monoid (All(..))
 import qualified Data.Text as Text
 import GHC.Stack (HasCallStack)
 
-#if MIN_VERSION_ghc(8,10,0)
+#if MIN_VERSION_ghc(9,6,0)
+import GHC.Core.Make (chunkify, mkChunkified)
+#elif MIN_VERSION_ghc(8,10,0)
 import GHC.Hs.Utils (chunkify, mkChunkified)
 #else
 import HsUtils (chunkify, mkChunkified)
@@ -88,13 +90,13 @@ import Clash.Core.Var (isGlobalId, isLocalId, varName)
 import Clash.Core.VarEnv
   ( InScopeSet, elemInScopeSet, extendInScopeSet, extendInScopeSetList
   , notElemInScopeSet, unionInScope)
+import qualified Clash.Data.UniqMap as UniqMap
 import Clash.Normalize.Transformations.Letrec (deadCode)
 import Clash.Normalize.Types (NormRewrite, NormalizeSession)
 import Clash.Rewrite.Combinators (bottomupR)
 import Clash.Rewrite.Types
 import Clash.Rewrite.Util (changed, isUntranslatableType)
 import Clash.Rewrite.WorkFree (isConstant)
-import Clash.Unique (lookupUniqMap)
 import Clash.Util (MonadUnique, curLoc)
 
 -- | This transformation lifts applications of global binders out of
@@ -301,7 +303,7 @@ collectGlobals' is0 substitution seen e@(collectArgsTicks -> (fun, args@(_:_), t
     let (ids1,ids2) = splitSupply ids
     uniqSupply Lens..= ids2
     gh <- Lens.use globalHeap
-    let eval = (Lens.view Lens._3) . whnf' evaluate bndrs tcm gh ids1 is0 False
+    let eval = (Lens.view Lens._3) . whnf' evaluate bndrs mempty tcm gh ids1 is0 False
     let eTy  = inferCoreTypeOf tcm e
     untran <- isUntranslatableType False eTy
     case untran of
@@ -478,7 +480,8 @@ mkDisjointGroup inScope (fun,(seen,cs)) = do
     let argss    = Foldable.toList cs
         argssT   = zip [0..] (List.transpose argss)
         (sharedT,distinctT) = List.partition (areShared tcm inScope . fmap (first stripTicks) . snd) argssT
-        shared   = map (second head) sharedT
+        -- TODO: find a better solution than "maybe undefined fst . uncons"
+        shared   = map (second (maybe (error "impossible") fst . List.uncons)) sharedT
         distinct = map (Either.lefts) (List.transpose (map snd distinctT))
         cs'      = fmap (zip [0..]) cs
         cs''     = removeEmpty
@@ -618,11 +621,12 @@ genCase tcm tupTcm ty argTys = go
 
 -- | Lookup the TyConName and DataCon for a tuple of size n
 findTup :: TyConMap -> IntMap TyConName -> Int -> (TyConName,DataCon)
-findTup tcm tupTcm n = (tupTcNm,tupDc)
-  where
-    tupTcNm      = Maybe.fromMaybe (error $ $curLoc ++ "Can't find " ++ show n ++ "-tuple") $ IntMap.lookup n tupTcm
-    Just tupTc   = lookupUniqMap tupTcNm tcm
-    [tupDc]      = tyConDataCons tupTc
+findTup tcm tupTcm n =
+  Maybe.fromMaybe (error ("Cannot build " <> show n <> "-tuble")) $ do
+    tupTcNm <- IntMap.lookup n tupTcm
+    tupTc <- UniqMap.lookup tupTcNm tcm
+    tupDc <- Maybe.listToMaybe (tyConDataCons tupTc)
+    return (tupTcNm,tupDc)
 
 mkBigTupTm :: TyConMap -> IntMap TyConName -> [(Type,Term)] -> Term
 mkBigTupTm tcm tupTcm args = snd $ mkBigTup tcm tupTcm args
@@ -763,7 +767,7 @@ interestingToLift inScope eval e@(Prim pInfo) args ticks
     allNonPow2  = all (not . termIsPow2) lArgs
     tailNonPow2 = case lArgs of
                     [] -> True
-                    _  -> all (not . termIsPow2) (tail lArgs)
+                    _  -> all (not . termIsPow2) (drop 1 lArgs)
     lastNotPow2 = case lArgs of
                     [] -> True
                     _  -> not (termIsPow2 (last lArgs))
