@@ -33,8 +33,9 @@ module Clash.Normalize.Transformations.Inline
 
 import qualified Control.Lens as Lens
 import qualified Control.Monad as Monad
+import Control.Monad ((>=>))
 import Control.Monad.Trans.Maybe (MaybeT(..))
-import Control.Monad.Writer ((>=>),lift,listen)
+import Control.Monad.Writer (lift,listen)
 import Data.Default (Default(..))
 import Data.Either  (lefts)
 import qualified Data.HashMap.Lazy as HashMap
@@ -44,12 +45,7 @@ import qualified Data.Monoid as Monoid (Any(..))
 import qualified Data.Text as Text
 import qualified Data.Text.Extra as Text
 import GHC.Stack (HasCallStack)
-
-#if MIN_VERSION_ghc(9,0,0)
-import GHC.Types.Basic             (InlineSpec (..))
-#else
-import BasicTypes                  (InlineSpec (..))
-#endif
+import GHC.BasicTypes.Extra (isNoInline)
 
 import qualified Clash.Explicit.SimIO as SimIO
 import qualified Clash.Sized.Internal.BitVector as BV (Bit(Bit), BitVector(BV))
@@ -464,12 +460,16 @@ inlineNonRep ctx0 e0@(Case {}) = do
   case r of
     (e1, Monoid.getAny -> True) ->
       return e1
-    (~(Case subj0 typ alts), _) -> do
+    (e1, _) -> do
       -- If a term _in_ the subject triggers 'inlineNonRepWorker', inline and
       -- propagate might eliminate this case. We therefore don't explore the
       -- alternatives. Note that this makes it substantially different from a
       -- 'topdownSucR' transformation.
       let
+        (subj0,typ,alts) = case e1 of
+          Case s t a -> (s,t,a)
+          _ -> error ("internal error, inlineNonRep triggered on a non-Case:" <>
+                      showPpr e1)
         TransformContext inScope ctx1 = ctx0
         ctx2 = TransformContext inScope (CaseScrut:ctx1)
 
@@ -602,7 +602,7 @@ inlineSmall _ e@(collectArgsTicks -> (Var f,args,ticks)) = do
         -- Don't inline recursive expressions
         Just b -> do
           isRecBndr <- isRecursiveBndr f
-          if not isRecBndr && bindingSpec b /= NoInline && termSize (bindingTerm b) < sizeLimit
+          if not isRecBndr && not (isNoInline (bindingSpec b)) && termSize (bindingTerm b) < sizeLimit
              then do
                let tm = mkTicks (bindingTerm b) (mkInlineTick f : ticks)
                changed $ mkApps tm args
