@@ -2,11 +2,11 @@
   Copyright  :  (C) 2012-2016, University of Twente,
                     2016-2017, Myrtle Software Ltd,
                     2017-2018, Google Inc.,
-                    2021-2022, QBayLogic B.V.
+                    2021-2023, QBayLogic B.V.
   License    :  BSD2 (see the file LICENSE)
   Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
 
-  Transformations for specialisation.
+  Transformations for specialization
 -}
 
 {-# LANGUAGE CPP #-}
@@ -80,6 +80,7 @@ import Clash.Core.Var (Var(..), Id, TyVar, mkTyVar)
 import Clash.Core.VarEnv
   ( InScopeSet, extendInScopeSet, extendInScopeSetList, lookupVarEnv
   , mkInScopeSet, mkVarSet, unionInScope, elemVarSet)
+import qualified Clash.Data.UniqMap as UniqMap
 import Clash.Debug (traceIf, traceM)
 import Clash.Driver.Types (Binding(..), TransformationInfo(..), hasTransformationInfo)
 import Clash.Netlist.Util (representableType)
@@ -95,9 +96,6 @@ import Clash.Normalize.Types
   ( NormRewrite, NormalizeSession, specialisationCache, specialisationHistory)
 import Clash.Normalize.Util
   (constantSpecInfo, csrFoundConstant, csrNewBindings, csrNewTerm)
-import Clash.Unique
-  ( eltsUniqMap, eltsUniqSet, extendUniqMapWith, unitUniqSet, filterUniqMap
-  , lookupUniqMap)
 import Clash.Util (ClashException(..))
 
 -- | Propagate arguments of application inwards; except for 'Lam' where the
@@ -127,10 +125,10 @@ import Clash.Util (ClashException(..))
 --       D a b -> h a b
 -- @
 --
--- is very bad because 'b' in 'h a b' is now bound by the pattern instead of the
+-- is very bad because @b@ in @h a b@ is now bound by the pattern instead of the
 -- newly introduced let-binding
 --
--- instead me must deshadow w.r.t. the new variable and rewrite to:
+-- instead we must deshadow w.r.t. the new variable and rewrite to:
 --
 -- @
 -- let b = f x y
@@ -158,8 +156,10 @@ import Clash.Util (ClashException(..))
 -- whether @x@ is in the current InScopeSet, and deshadow if that's the case,
 -- i.e. we then rewrite to:
 --
+-- @
 -- let x1 = u
 -- in  e [x:=x1]
+-- @
 --
 -- Case 3.
 --
@@ -282,7 +282,7 @@ appProp ctx@(TransformContext is _) = \case
   goCaseArg _ ty ls [] = return (ty,ls,[])
 {-# SCC appProp #-}
 
--- | Specialise functions on arguments which are constant, except when they
+-- | Specialize functions on arguments which are constant, except when they
 -- are clock, reset generators.
 constantSpec :: HasCallStack => NormRewrite
 constantSpec ctx@(TransformContext is0 tfCtx) e@(App e1 e2)
@@ -341,6 +341,27 @@ We may also not want to look at ticks, as then the specialization cache will
 miss on virtually every lookup which could add to normalization time.
 -}
 
+-- | Given two 'InlineSpec's, return the \"strongest\" one. I.e., the one that's
+-- closest to @NoInline@ (or @Opaque@ for newer GHCs).
+preferNoInline :: InlineSpec -> InlineSpec -> InlineSpec
+preferNoInline is0 is1
+  | enumInlineSpec is0 >= enumInlineSpec is1 = is0
+  | otherwise                                = is1
+ where
+  enumInlineSpec :: InlineSpec -> Int
+  enumInlineSpec = \case
+#if MIN_VERSION_ghc(9,2,0)
+    NoUserInlinePrag {} -> -1
+#else
+    NoUserInline {} -> -1
+#endif
+    Inline {} -> 0
+    Inlinable {} -> 1
+    NoInline {} -> 2
+#if MIN_VERSION_ghc(9,4,0)
+    Opaque {} -> 3
+#endif
+
 -- | Specialize an application on its argument
 specialize'
   :: TransformContext
@@ -398,11 +419,11 @@ specialize' (TransformContext is0 _) e (Var f, args, ticks) specArgIn = do
     -- Create new specialized function
     Nothing -> do
       -- Determine if we can specialize f
-      bodyMaybe <- fmap (lookupUniqMap (varName f)) $ Lens.use bindings
+      bodyMaybe <- fmap (UniqMap.lookup f) $ Lens.use bindings
       case bodyMaybe of
         Just (Binding _ sp inl _ bodyTm _) -> do
           -- Determine if we see a sequence of specializations on a growing argument
-          specHistM <- lookupUniqMap f <$> Lens.use (extra.specialisationHistory)
+          specHistM <- UniqMap.lookup f <$> Lens.use (extra.specialisationHistory)
           specLim   <- Lens.view specializationLimit
           if maybe False (> specLim) specHistM
             then throw (ClashException
@@ -430,7 +451,7 @@ specialize' (TransformContext is0 _) e (Var f, args, ticks) specArgIn = do
               (fId,inl',specArg') <- case specArg of
                 Left a@(collectArgsTicks -> (Var g,gArgs,_gTicks)) -> if isPolyFun tcm a
                     then do
-                      -- In case we are specialising on an argument that is a
+                      -- In case we are specializing on an argument that is a
                       -- global function then we use that function's name as the
                       -- name of the specialized higher-order function.
                       -- Additionally, we will return the body of the global
@@ -441,21 +462,32 @@ specialize' (TransformContext is0 _) e (Var f, args, ticks) specArgIn = do
                       -- binding @g'@ where both the body of @mealy@ and @g@
                       -- are inlined, meaning the state-transition-function
                       -- and the memory element will be in a single function.
-                      gTmM <- fmap (lookupUniqMap (varName g)) $ Lens.use bindings
-                      return (g,maybe inl bindingSpec gTmM, maybe specArg (Left . (`mkApps` gArgs) . bindingTerm) gTmM)
+                      gTmM <- fmap (UniqMap.lookup g) $ Lens.use bindings
+                      return
+                        ( g
+                        , preferNoInline inl (maybe noUserInline bindingSpec gTmM)
+                        , maybe specArg (Left . (`mkApps` gArgs) . bindingTerm) gTmM
+                        )
                     else return (f,inl,specArg)
                 _ -> return (f,inl,specArg)
               -- Create specialized functions
               let newBody = mkAbstraction (mkApps bodyTm (argVars ++ [specArg'])) (boundArgs ++ specBndrs)
               newf <- mkFunction (varName fId) sp inl' newBody
               -- Remember specialization
-              (extra.specialisationHistory) %= extendUniqMapWith f 1 (+)
+              (extra.specialisationHistory) %= UniqMap.insertWith (+) f 1
               (extra.specialisationCache)  %= Map.insert (f,argLen,specAbs) newf
               -- use specialized function
               let newExpr = mkApps (mkTicks (Var newf) ticks) (args ++ specVars)
               newf `deepseq` changed newExpr
         Nothing -> return e
   where
+    noUserInline :: InlineSpec
+#if MIN_VERSION_ghc(9,2,0)
+    noUserInline = NoUserInlinePrag
+#else
+    noUserInline = NoUserInline
+#endif
+
     collectBndrsMinusApps :: Term -> [Name a]
     collectBndrsMinusApps = reverse . go []
       where
@@ -469,7 +501,7 @@ specialize' (TransformContext is0 _) e (Var f, args, ticks) specArgIn = do
           bs' -> init bs' ++ bs
         go bs _ = bs
 
--- Specialising non Var's is used by nonRepANF
+-- Specializing non Var's is used by nonRepANF
 specialize' _ctx _ (appE,args,ticks) (Left specArg) = do
   -- Create binders and variable references for free variables in 'specArg'
   let (specBndrs,specVars) = specArgBndrsAndVars (Left specArg)
@@ -477,11 +509,15 @@ specialize' _ctx _ (appE,args,ticks) (Left specArg) = do
       newBody = mkAbstraction specArg specBndrs
   -- See if there's an existing binder that's alpha-equivalent to the
   -- specialized function
-  existing <- filterUniqMap ((`aeqTerm` newBody) . bindingTerm) <$> Lens.use bindings
+  existing <- UniqMap.filter ((`aeqTerm` newBody) . bindingTerm) <$> Lens.use bindings
   -- Create a new function if an alpha-equivalent binder doesn't exist
-  newf <- case eltsUniqMap existing of
+  newf <- case UniqMap.elems existing of
     [] -> do (cf,sp) <- Lens.use curFun
+#if MIN_VERSION_ghc(9,2,0)
+             mkFunction (appendToName (varName cf) "_specF") sp NoUserInlinePrag newBody
+#else
              mkFunction (appendToName (varName cf) "_specF") sp NoUserInline newBody
+#endif
     (b:_) -> return (bindingId b)
   -- Create specialized argument
   let newArg  = Left $ mkApps (Var newf) specVars
@@ -497,10 +533,10 @@ specialize' _ e _ _ = return e
 -- yield (alpha equivalent) results for the same specialization. While collecting
 -- free variables in a given term or type it should therefore keep a stable
 -- ordering based on the order in which it finds free vars. To see why,
--- consider the following two pseudo-code calls to 'specialise':
+-- consider the following two pseudo-code calls to 'specialize':
 --
---     specialise {f ('a', x[123], y[456])}
---     specialise {f ('b', x[456], y[123])}
+--     specialize {f ('a', x[123], y[456])}
+--     specialize {f ('b', x[456], y[123])}
 --
 -- Collecting the binders in a VarSet would yield the following (unique ordered)
 -- sets:
@@ -528,7 +564,9 @@ specArgBndrsAndVars specArg =
       (specFTVs,specFVs) = case specArg of
         Left tm  -> (OSet.toListL *** OSet.toListL) . getConst $
                     Lens.foldMapOf freeLocalVars unitFV tm
-        Right ty -> (eltsUniqSet (Lens.foldMapOf typeFreeVars unitUniqSet ty),[] :: [Id])
+        Right ty -> ( UniqMap.elems
+                        (Lens.foldMapOf typeFreeVars (\x -> UniqMap.singletonUnique (coerce x)) ty)
+                    , [] :: [Id])
 
       specTyBndrs = map Right specFTVs
       specTmBndrs = map Left  specFVs
@@ -558,12 +596,12 @@ nonRepSpec ctx e@(App e1 e2)
            specialize ctx (App e1 e2')
          else return e
   where
-    -- | If the argument on which we're specialising ia an internal function,
+    -- | If the argument on which we're specializing is an internal function,
     -- one created by the compiler, then inline that function before we
-    -- specialise.
+    -- specialize.
     --
-    -- We need to do this because otherwise the specialisation history won't
-    -- recognize the new specialisation argument as something the function has
+    -- We need to do this because otherwise the specialization history won't
+    -- recognize the new specialization argument as something the function has
     -- already been specialized on
     inlineInternalSpecialisationArgument
       :: Term
@@ -596,8 +634,8 @@ typeSpec _ e = return e
 {-# SCC typeSpec #-}
 
 -- | Specialize functions on arguments which are zero-width. These arguments
--- can have only one possible value, and specialising on this value may create
--- additional oppourtunities for transformations to fire.
+-- can have only one possible value, and specializing on this value may create
+-- additional opportunities for transformations to fire.
 --
 -- As we can't remove zero-width arguements (as transformations cannot change
 -- the type of a term), we instead substitute all occurances of a lambda-bound
@@ -646,7 +684,7 @@ zeroWidthTypeElem tcm ty = do
      -- where all fields are also zero-width.
      | otherwise
      -> do
-       tc <- lookupUniqMap tcNm tcm
+       tc <- UniqMap.lookup tcNm tcm
 
        case tyConDataCons tc of
          [dc] -> do
