@@ -1,7 +1,8 @@
 {-|
-  Copyright   :  (C) 2019, Myrtle Software Ltd.
-                     2020-2022, QBayLogic B.V.
-                     2021, Myrtle.ai
+  Copyright   :  (C) 2019,      Myrtle Software Ltd.
+                     2020-2023, QBayLogic B.V.
+                     2021,      Myrtle.ai
+                     2022-2023, Google Inc
   License     :  BSD2 (see the file LICENSE)
   Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -9,8 +10,11 @@ This module contains a mini dsl for creating haskell blackbox
 instantiations.
 -}
 
+{-# LANGUAGE CPP               #-}
 {-# LANGUAGE LambdaCase        #-}
+{-# LANGUAGE NamedFieldPuns    #-}
 {-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE NamedFieldPuns    #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE PatternSynonyms   #-}
 {-# LANGUAGE QuasiQuotes       #-}
@@ -27,9 +31,14 @@ module Clash.Primitives.DSL
 
   -- * Declarations
   , BlockState (..)
-  , TExpr
+  , TExpr(..)
+  , addDeclaration
+  , assign
+  , compInBlock
   , declaration
   , declarationReturn
+  , declare
+  , declareN
   , instDecl
   , instHO
   , viaAnnotatedSignal
@@ -48,14 +57,18 @@ module Clash.Primitives.DSL
   , tResults
   , getStr
   , getBool
+  , getVec
   , exprToInteger
   , tExprToInteger
   , deconstructProduct
   , untuple
   , unvec
+  , deconstructMaybe
 
   -- ** Conversion
+  , bitCoerce
   , toBV
+  , toBvWithAttrs
   , fromBV
   , enableToBit
   , boolToBit
@@ -64,6 +77,9 @@ module Clash.Primitives.DSL
   , unsignedFromBitVector
   , boolFromBits
 
+  , unsafeToActiveHigh
+  , unsafeToActiveLow
+
   -- ** Operations
   , andExpr
   , notExpr
@@ -72,12 +88,16 @@ module Clash.Primitives.DSL
   , open
 
   -- ** Utilities
+  , clog2
+  , litTExpr
   , toIdentifier
   , tySize
-  , clog2
   ) where
 
 import           Control.Lens                    hiding (Indexed, assign)
+#if MIN_VERSION_mtl(2,3,0)
+import           Control.Monad                   (forM, forM_, zipWithM)
+#endif
 import           Control.Monad.State
 import           Data.Default                    (Default(def))
 import           Data.IntMap                     (IntMap)
@@ -95,10 +115,11 @@ import           Data.Text.Prettyprint.Doc.Extra
 import           GHC.Stack                       (HasCallStack)
 
 import           Clash.Annotations.Primitive     (HDL (..), Primitive (..))
-import           Clash.Backend                   hiding (fromBV, toBV)
+import           Clash.Annotations.SynthesisAttributes (Attr)
+import           Clash.Backend                   hiding (Usage, fromBV, toBV)
 import           Clash.Backend.VHDL              (VHDLState)
-import           Clash.Core.Var                  (Attr')
-import           Clash.Netlist.BlackBox.Util     (exprToString, renderElem)
+import           Clash.Explicit.Signal           (ResetPolarity(..), vResetPolarity)
+import           Clash.Netlist.BlackBox.Util     (exprToString, getDomainConf, renderElem)
 import           Clash.Netlist.BlackBox.Types
   (BlackBoxTemplate, Element(Component, Text), Decl(..))
 import qualified Clash.Netlist.Id                as Id
@@ -106,7 +127,6 @@ import           Clash.Netlist.Types             hiding (Component, toBit)
 import           Clash.Netlist.Util
 import           Clash.Util                      (clogBase)
 import qualified Data.String.Interpolate         as I
-import           Data.String.Interpolate.Util    (unindent)
 import           Language.Haskell.TH             (Name)
 import           Prelude
 
@@ -139,7 +159,7 @@ instance Default BlackBoxHaskellOpts where
 -- | Create a blackBoxHaskell primitive. To be used as part of an annotation:
 --
 -- @
--- {-\# ANN myFunction (blackBoxHaskell 'myFunction 'myBBF def{_ignoredArguments=[1,2]}) \#-}
+-- {-\# ANN myFunction (blackBoxHaskell 'myFunction 'myBBF def{bo_ignoredArguments=[1,2]}) \#-}
 -- @
 --
 -- @[1,2]@ would mean this blackbox __ignores__ its second and third argument.
@@ -152,19 +172,17 @@ blackBoxHaskell
   -- ^ Options, see data structure for more information
   -> Primitive
 blackBoxHaskell bb tf BlackBoxHaskellOpts{..} =
-  InlinePrimitive bo_supportedHdls $ unindent [I.i|
-  [ { "BlackBoxHaskell" :
-      { "name" : "#{bb}"
-      , "templateFunction" : "#{tf}"
-      , "ignoredArguments" : #{show bo_ignoredArguments}
-      , "multiResult" : #{toJsonBool bo_multiResult}
-      }
-    }
-  ] |]
+  InlineYamlPrimitive bo_supportedHdls [I.__i|
+    BlackBoxHaskell:
+      name: #{bb}
+      templateFunction: #{tf}
+      ignoredArguments : #{bo_ignoredArguments}
+      multiResult : #{toYamlBool bo_multiResult}
+    |]
  where
-  toJsonBool :: Bool -> String
-  toJsonBool True = "true"
-  toJsonBool False = "false"
+  toYamlBool :: Bool -> String
+  toYamlBool True = "true"
+  toYamlBool False = "false"
 
 -- | The state of a block. Contains a list of declarations and a the
 --   backend state.
@@ -183,6 +201,20 @@ instance Backend backend => HasIdentifierSet (BlockState backend) where
   identifierSet :: Lens' (BlockState backend) IdentifierSet
   identifierSet = bsBackend . identifierSet
 
+instance HasUsageMap backend => HasUsageMap (BlockState backend) where
+  usageMap = bsBackend.usageMap
+
+liftToBlockState
+  :: forall backend a. Backend backend
+   => State backend a -> State (BlockState backend) a
+liftToBlockState (StateT f) = StateT g
+ where
+  g :: BlockState backend -> Identity (a, BlockState backend)
+  g sbsIn = do
+    let sIn = _bsBackend sbsIn
+    (res,sOut) <- f sIn
+    pure (res, sbsIn{_bsBackend = sOut})
+
 -- | A typed expression.
 data TExpr = TExpr
   { ety :: HWType
@@ -205,9 +237,18 @@ declarationReturn
 declarationReturn bbCtx blockName blockBuilder =
   declaration blockName $ do
     res <- blockBuilder
-    forM_ (zip (bbResults bbCtx) res) $ \(rNm, r) -> do
-      let (Identifier resultNm Nothing, _) = rNm
-      addDeclaration (Assignment resultNm (eex r))
+    forM_ (zip (bbResults bbCtx) res) $ \(rNm, r) -> case rNm of
+      (Identifier resultNm Nothing, _) ->
+        addDeclaration (Assignment resultNm Cont (eex r))
+      (t,_) -> error ("declarationReturn expected an Identifier, but got: " <> show t)
+
+
+emptyBlockState :: backend -> BlockState backend
+emptyBlockState bck = BlockState
+  { _bsDeclarations = []
+  , _bsHigherOrderCalls = IntMap.empty
+  , _bsBackend = bck
+  }
 
 -- | Run a block declaration.
 declaration
@@ -218,13 +259,13 @@ declaration
   -- ^ block builder
   -> State backend Doc
   -- ^ pretty printed block
-declaration blockName s = do
+declaration blockName c = do
   backend0 <- get
-  let initState = BlockState [] IntMap.empty backend0
-      BlockState decs _hoCalls backend1 = execState s initState
-  put backend1
+  let initState = emptyBlockState backend0
+      (BlockState {..}) = execState c initState
+  put _bsBackend
   blockNameUnique <- Id.makeBasic blockName
-  getAp $ blockDecl blockNameUnique (reverse decs)
+  getAp $ blockDecl blockNameUnique (reverse _bsDeclarations)
 
 -- | Add a declaration to the state.
 addDeclaration :: Declaration -> State (BlockState backend) ()
@@ -235,15 +276,13 @@ declare'
   :: Backend backend
   => Text
   -- ^ Name hint
-  -> WireOrReg
-  -- ^ Should signal be declared as a wire or a reg
   -> HWType
   -- ^ Type of new signal
   -> State (BlockState backend) Identifier
   -- ^ Expression pointing the the new signal
-declare' decName wireOrReg ty = do
+declare' decName ty = do
   uniqueName <- Id.makeBasic decName
-  addDeclaration (NetDecl' Nothing wireOrReg uniqueName (Right ty) Nothing)
+  addDeclaration (NetDecl' Nothing uniqueName ty Nothing)
   pure uniqueName
 
 -- | Declare a new signal with the given name and type.
@@ -251,16 +290,33 @@ declare
   :: Backend backend
   => Text
   -- ^ Name hint
-  -> WireOrReg
-  -- ^ Should signal be declared as a wire or a reg
   -> HWType
   -- ^ Type of new signal
   -> State (BlockState backend) TExpr
   -- ^ Expression pointing the the new signal
-declare decName wireOrReg ty = do
-  uniqueName <- declare' decName wireOrReg ty
+declare decName ty = do
+  uniqueName <- declare' decName ty
   pure (TExpr ty (Identifier uniqueName Nothing))
 
+-- | Declare /n/ new signals with the given type and based on the given name
+declareN
+  :: Backend backend
+  => Text
+  -- ^ Name hint
+  -> [HWType]
+  -- ^ Types of the signals
+  -> State (BlockState backend) [TExpr]
+  -- ^ Expressions pointing the the new signals
+declareN decName tys = do
+  firstName <- Id.makeBasic decName
+  nextNames <- Id.nextN (length tys - 1) firstName
+  let uniqueNames = firstName : nextNames
+  zipWithM
+    (\uniqueName ty -> do
+      addDeclaration $ NetDecl' Nothing uniqueName ty Nothing
+      pure $ TExpr ty (Identifier uniqueName Nothing)
+    ) uniqueNames tys
+
 -- | Assign an expression to an identifier, returns the new typed
 --   identifier expression.
 assign
@@ -268,12 +324,16 @@ assign
   => Text
   -- ^ Name hint for assignment
   -> TExpr
-  -- ^ expression to be assigned to
+  -- ^ expression to be assigned to freshly generated identifier
   -> State (BlockState backend) TExpr
   -- ^ the identifier of the expression that actually got assigned
 assign aName (TExpr ty aExpr) = do
-  texp@(~(TExpr _ (Identifier uniqueName Nothing))) <- declare aName Wire ty
-  addDeclaration (Assignment uniqueName aExpr)
+  texp <- declare aName ty
+  let uniqueName = case texp of
+        TExpr _ (Identifier x Nothing) -> x
+        t' -> error ("assign expected an Identifier, but got: " <> show t')
+
+  addDeclaration (Assignment uniqueName Cont aExpr)
   pure texp
 
 -- | Extract the elements of a vector expression and return expressions
@@ -288,11 +348,49 @@ unvec
   -> State (BlockState backend) [TExpr]
   -- ^ Vector elements
 unvec vName v@(ety -> Vector vSize eType) = do
-  ~(TExpr _ (Identifier vUniqueName Nothing)) <- toIdentifier vName v
+  texp <- toIdentifier vName v
+  let vUniqueName = case texp of
+        TExpr _ (Identifier x Nothing) -> x
+        t' -> error ("unvec expected an Identifier, but got: " <> show t')
+
   let vIndex i = Identifier vUniqueName (Just (Indexed (ety v, 10, i)))
   pure (map (TExpr eType . vIndex) [0..vSize-1])
 unvec _ e = error $ "unvec: cannot be called on non-vector: " <> show (ety e)
 
+-- | Deconstruct a 'Maybe' into its constructor 'Bit' and contents of its 'Just'
+-- field. Note that the contents might be undefined, if the constructor bit is
+-- set to 'Nothing'.
+deconstructMaybe ::
+  (HasCallStack, Backend backend) =>
+  -- | Maybe expression
+  TExpr ->
+  -- | Name hint for constructor bit, data
+  (Text, Text) ->
+  -- | Constructor represented as a Bit, contents of Just
+  State (BlockState backend) (TExpr, TExpr)
+deconstructMaybe e@TExpr{ety} (bitName, contentName)
+  | SP tyName [(_nothing, []),(_just, [aTy])] <- ety
+  , tyName == fromString (show ''Maybe)
+  = do
+    eBv <- toBV (bitName <> "_and_" <> contentName <> "_bv") e
+    eId <- toIdentifier' (bitName <> "_and_" <> contentName) eBv
+    let eSize = typeSize ety
+
+    bitExpr <- fromBV bitName Bit TExpr
+      { eex = Identifier eId (Just (Sliced (BitVector eSize, eSize - 1, eSize - 1)))
+      , ety = BitVector 1
+      }
+
+    contentExpr <- fromBV contentName aTy TExpr
+      { eex = Identifier eId (Just (Sliced (BitVector eSize, eSize - 1 - 1, 0)))
+      , ety = BitVector (eSize - 1)
+      }
+
+    pure (bitExpr, contentExpr)
+
+deconstructMaybe e _ =
+  error $ "deconstructMaybe: cannot be called on non-Maybe: " <> show (ety e)
+
 -- | Extract the fields of a product type and return expressions
 --   to them. These new expressions are given unique names and get
 --   declared in the block scope.
@@ -303,10 +401,15 @@ deconstructProduct
   -> [Text]
   -- ^ Name hints for element assignments
   -> State (BlockState backend) [TExpr]
-deconstructProduct (TExpr ty@(Product _ _ tys) (Identifier resName _)) vals = do
-  newNames <- zipWithM (flip declare Wire) vals tys
-  addDeclaration $ Assignment resName $ DataCon ty (DC (ty, 0)) (fmap eex newNames)
-  pure newNames
+deconstructProduct (TExpr ty@(Product _ _ fieldTys) (Identifier resName Nothing)) nameHints =
+  forM (zip3 [0..] nameHints fieldTys) $ \(fieldIndex, nameHint, fieldTy) ->
+    assign nameHint $
+      TExpr fieldTy (Identifier resName (Just (Indexed (ty, 0, fieldIndex))))
+
+deconstructProduct t0@(TExpr (Product {}) _) nameHints = do
+  t1 <- toIdentifier "product" t0
+  deconstructProduct t1 nameHints
+
 deconstructProduct e i =
   error $ "deconstructProduct: " <> show e <> " " <> show i
 
@@ -365,12 +468,16 @@ boolToBit bitName = \case
   T -> pure High
   F -> pure Low
   TExpr Bool boolExpr -> do
-    texp@(~(TExpr _ (Identifier uniqueBitName Nothing))) <- declare bitName Wire Bit
+    texp <- declare bitName Bit
+    let uniqueBitName = case texp of
+          TExpr _ (Identifier x Nothing) -> x
+          t' -> error ("boolFromBit expected an Identifier, but got: " <> show t')
     addDeclaration $
       CondAssignment uniqueBitName Bit boolExpr Bool
         [ (Just (BoolLit True), Literal Nothing (BitLit H))
         , (Nothing            , Literal Nothing (BitLit L))
         ]
+    declareUseOnce (Proc NonBlocking) uniqueBitName
     pure texp
   tExpr -> error $ "boolToBit: Got \"" <> show tExpr <> "\" expected Bool"
 
@@ -383,13 +490,17 @@ enableToBit
   -> State (BlockState backend) TExpr
 enableToBit bitName = \case
   TExpr ena@(Enable _) enableExpr -> do
-    texp@(~(TExpr _ (Identifier uniqueBitName Nothing))) <- declare bitName Wire Bit
+    texp <- declare bitName Bit
+    let uniqueBitName = case texp of
+          TExpr _ (Identifier x Nothing) -> x
+          t' -> error ("boolFromBit expected an Identifier, but got: " <> show t')
     addDeclaration $
       CondAssignment uniqueBitName Bit enableExpr ena
         -- Enable normalizes to Bool for all current backends
         [ (Just (BoolLit True), Literal Nothing (BitLit H))
         , (Nothing            , Literal Nothing (BitLit L))
         ]
+    declareUseOnce (Proc NonBlocking) uniqueBitName
     pure texp
   tExpr -> error $ "enableToBit: Got \"" <> show tExpr <> "\" expected Enable"
 
@@ -398,11 +509,27 @@ enableToBit bitName = \case
 --   Returns a reference to a declared `Bit` that should get assigned
 --   by something (usually the output port of an entity).
 boolFromBit
-  :: Text
+  :: (HasCallStack, Backend backend)
+  => Text
   -- ^ Name hint for intermediate signal
   -> TExpr
-  -> State (BlockState VHDLState) TExpr
-boolFromBit = outputCoerce Bit Bool (<> " = '1'")
+  -> State (BlockState backend) TExpr
+boolFromBit boolName = \case
+  High -> pure T
+  Low -> pure F
+  TExpr Bit bitExpr -> do
+    texp <- declare boolName Bool
+    let uniqueBoolName = case texp of
+          TExpr _ (Identifier x Nothing) -> x
+          t' -> error ("boolFromBit expected an Identifier, but got: " <> show t')
+    addDeclaration $
+      CondAssignment uniqueBoolName Bool bitExpr Bit
+        [ (Just (BitLit H), Literal Nothing (BoolLit True))
+        , (Nothing        , Literal Nothing (BoolLit False))
+        ]
+    declareUseOnce (Proc NonBlocking) uniqueBoolName
+    pure texp
+  tExpr -> error $ "boolFromBit: Got \"" <> show tExpr <> "\" expected Bit"
 
 -- | Used to create an output `Bool` from a `BitVector` of given size.
 -- Works in a similar way to `boolFromBit` above.
@@ -419,17 +546,18 @@ boolFromBitVector n =
 
 -- | Used to create an output `Unsigned` from a `BitVector` of given
 -- size. Works in a similar way to `boolFromBit` above.
---
--- TODO: Implement for (System)Verilog
-unsignedFromBitVector
-  :: HasCallStack
-  => Size
-  -> Text
-  -- ^ Name hint for intermediate signal
-  -> TExpr
-  -> State (BlockState VHDLState) TExpr
-unsignedFromBitVector n =
-  outputCoerce (BitVector n) (Unsigned n) (\i -> "unsigned(" <> i <> ")")
+unsignedFromBitVector ::
+  (HasCallStack, Backend backend) =>
+  -- | Name hint for intermediate signal
+  Text ->
+  -- | BitVector expression
+  TExpr ->
+  -- | Unsigned expression
+  State (BlockState backend) TExpr
+unsignedFromBitVector nameHint e@TExpr{ety=BitVector n} =
+  fromBV nameHint (Unsigned n) e
+unsignedFromBitVector _nameHint TExpr{ety} =
+  error $ "unsignedFromBitVector: Expected BitVector, got: " <> show ety
 
 -- | Used to create an output `Bool` from a number of `Bit`s, using
 -- conjunction. Similarly to `untuple`, it returns a list of
@@ -465,7 +593,7 @@ outputCoerce fromType toType exprStringFn inName0 expr_
       let inName2 = Id.unsafeMake (exprStringFn (Id.toText inName1))
           exprIdent = Identifier inName2 Nothing
       addDeclaration (NetDecl Nothing inName1 fromType)
-      addDeclaration (Assignment outName exprIdent)
+      addDeclaration (Assignment outName Cont exprIdent)
       pure (TExpr fromType (Identifier inName1 Nothing))
 outputCoerce _ toType _ _ texpr = error $ "outputCoerce: the expression " <> show texpr
                                   <> " must be an Identifier with type " <> show toType
@@ -491,7 +619,7 @@ outputFn fromTypes toType exprFn inNames0 (TExpr outType (Identifier outName Not
           exprIdent = Identifier idExpr Nothing
       sequenceOf_ each [ addDeclaration (NetDecl Nothing nm t)
                        | (nm, t) <- zip inNames1 fromTypes ]
-      addDeclaration (Assignment outName exprIdent)
+      addDeclaration (Assignment outName Cont exprIdent)
       pure [ TExpr t (Identifier nm Nothing)
            | (nm,t) <- zipEqual inNames1 fromTypes ]
 outputFn _ outType _ _ texpr =
@@ -539,6 +667,17 @@ getBool :: TExpr -> Maybe Bool
 getBool (TExpr _ (Literal _ (BoolLit b))) = Just b
 getBool _ = Nothing
 
+-- | Try to get a Vector of expressions.
+getVec :: TExpr -> Maybe [TExpr]
+getVec (TExpr (Void (Just (Vector 0 _) )) _) =
+  pure []
+getVec (TExpr (Vector 1 elementTy) (DataCon _ VecAppend [e])) =
+  pure [TExpr elementTy e]
+getVec (TExpr (Vector n elementTy) (DataCon _ VecAppend [e, es0])) = do
+  es1 <- getVec (TExpr (Vector (n-1) elementTy) es0)
+  pure (TExpr elementTy e:es1)
+getVec _ = Nothing
+
 -- | Try to get the literal nat value of an expression.
 tExprToInteger :: TExpr -> Maybe Integer
 tExprToInteger (TExpr _ e) = exprToInteger e
@@ -548,41 +687,72 @@ exprToInteger (DataCon _ _ [n]) = exprToInteger n
 exprToInteger (Literal _ (NumLit n)) = Just n
 exprToInteger _ = Nothing
 
--- | Assign an input bitvector to an expression. Declares a new
---   bitvector if the expression is not already a bitvector.
-toBV
-  :: Backend backend
-  => Text
-  -- ^ BitVector name hint
-  -> TExpr
-  -- ^ expression
-  -> State (BlockState backend) TExpr
-  -- ^ BitVector expression
-toBV bvName a = case a of
-  TExpr BitVector{} _ -> pure a
-  TExpr aTy aExpr     -> assign bvName $
-    TExpr (BitVector (typeSize aTy)) (ToBv Nothing aTy aExpr)
+-- | Convert an expression from one type to another. Errors if result type and
+-- given expression are sized differently.
+bitCoerce ::
+  (HasCallStack, Backend backend) =>
+  -- | Name hints for intermediate variables
+  Text ->
+  -- | Type to convert to
+  HWType ->
+  -- | Expression to convert
+  TExpr ->
+  -- | Converted expression
+  State (BlockState backend) TExpr
+bitCoerce nameHint destType e@(TExpr ety _)
+  | tySize ety /= tySize @Int destType = error "Size mismatch"
+  | ety == destType = pure e
+  | BitVector _ <- ety = fromBV nameHint destType e
+  | otherwise = bitCoerce nameHint destType =<< toBV nameHint e
 
--- | Assign an output bitvector to an expression. Declares a new
---   bitvector if the expression is not already a bitvector.
+-- | Convert an expression to a BitVector
+toBV ::
+  Backend backend =>
+  -- | BitVector name hint
+  Text ->
+  -- | Expression to convert to BitVector
+  TExpr ->
+  -- | BitVector expression
+  State (BlockState backend) TExpr
+toBV = toBvWithAttrs []
+
+-- | Convert an expression to a BitVector and add the given HDL attributes
+toBvWithAttrs ::
+  Backend backend =>
+  [Attr Text] ->
+  -- | BitVector name hint
+  Text ->
+  -- | Expression to convert to BitVector
+  TExpr ->
+  -- | BitVector expression
+  State (BlockState backend) TExpr
+toBvWithAttrs attrs bvName (TExpr aTy aExpr) =
+  assign bvName $
+    TExpr
+      (annotated attrs (BitVector (tySize aTy)))
+      (ToBv Nothing aTy aExpr)
+
+-- | Convert an expression from a 'BitVector' into some type. If the expression
+-- is 'Annotated', only convert the expression within.
 fromBV
-  :: (HasCallStack, Backend backend)
-  => Text
-  -- ^ BitVector name hint
-  -> TExpr
-  -- ^ expression
-  -> State (BlockState backend) TExpr
-  -- ^ bv expression
-fromBV _ a@(TExpr BitVector{} _) = pure a
-fromBV bvName (TExpr aTy (Identifier aName Nothing)) = do
-  bvName' <- Id.makeBasic bvName
-  let bvExpr = FromBv Nothing aTy (Identifier bvName' Nothing)
-      bvTy   = BitVector (typeSize aTy)
-  addDeclaration (NetDecl Nothing bvName' bvTy)
-  addDeclaration (Assignment aName bvExpr)
-  pure (TExpr bvTy (Identifier bvName' Nothing))
-fromBV _ texpr = error $
-  "fromBV: the expression " <> show texpr <> "must be an Identifier"
+  :: (HasCallStack, Backend backend) =>
+  -- | Result name hint
+  Text ->
+  -- | Type to convert to
+  HWType ->
+  -- | 'BitVector' expression
+  TExpr ->
+  -- | Converted 'BitVector' expression
+  State (BlockState backend) TExpr
+fromBV resultName resultType e@TExpr{eex, ety = BitVector _} =
+  case resultType of
+    BitVector{} -> pure e
+    _ -> assign resultName (TExpr resultType (FromBv Nothing resultType eex))
+fromBV resultName resultType e@TExpr{ety = Annotated _ bv@(BitVector _)} =
+  case resultType of
+    BitVector{} -> pure (TExpr bv (eex e))
+    _ -> assign resultName (TExpr resultType (FromBv Nothing resultType (eex e)))
+fromBV _ _ TExpr{ety} = error $ "fromBV: expected BitVector, got: " <> show ety
 
 clog2 :: Num i => Integer -> i
 clog2 = fromIntegral . fromMaybe 0 . clogBase 2
@@ -618,11 +788,11 @@ instHO
   -> Int
   -- ^ Position of HO-argument. For example:
   --
-  --   fold :: forall n a . (a -> a -> a) -> Vec (n + 1) a -> a
+  -- > fold :: forall n a . (a -> a -> a) -> Vec (n + 1) a -> a
   --
   -- would have its HO-argument at position 0, while
   --
-  --  iterateI :: forall n a. KnownNat n => (a -> a) -> a -> Vec n a
+  -- > iterateI :: forall n a. KnownNat n => (a -> a) -> a -> Vec n a
   --
   -- would have it at position 1.
   -> (HWType, BlackBoxTemplate)
@@ -646,15 +816,8 @@ instHO bbCtx fPos (resTy, bbResTy) argsWithTypes = do
   let
     args2 = map (pure . Text . Id.toLazyText) args1
 
-    -- Create result identifier
-    -- See https://github.com/clash-lang/clash-compiler/issues/919 for info on
-    -- logic of 'resWireOrReg'
-    resWireOrReg =
-      case IntMap.lookup fPos (bbFunctions bbCtx) of
-        Just ((_,rw,_,_,_,_):_) -> rw
-        _ -> error "internal error"
   resName <- declare' (ctxName <> "_" <> "ho" <> showt fPos <> "_"
-                               <> showt fSubPos <> "_res") resWireOrReg resTy
+                               <> showt fSubPos <> "_res") resTy
   let res = ([Text (Id.toLazyText resName)], bbResTy)
 
   -- Render HO argument to plain text
@@ -674,33 +837,76 @@ instHO bbCtx fPos (resTy, bbResTy) argsWithTypes = do
 
   pure (TExpr resTy (Identifier resName Nothing))
 
--- | Instantiate a component/entity in a block state.
+-- | This creates a component declaration (for VHDL) given in and out port
+-- names, updating the 'BlockState backend' stored in the 'State' monad.
+--
+-- A typical result is that a
+--
+-- @
+-- component fifo port
+--    ( rst : in std_logic
+--    ...
+--    ; full : out std_logic
+--    ; empty : out std_logic );
+--  end component;
+-- @
+--
+-- declaration would be added in the appropriate place.
+compInBlock
+  :: forall backend
+   . Backend backend
+  => Text
+  -- ^ Component name
+  -> [(Text, HWType)]
+  -- ^ in ports
+  -> [(Text, HWType)]
+  -- ^ out ports
+  -> State (BlockState backend) ()
+compInBlock compName inPorts0 outPorts0 =
+  addDeclaration (CompDecl compName (inPorts1 ++ outPorts1))
+ where
+  mkPort inOut (nm, ty) = (nm, inOut, ty)
+  inPorts1 = mkPort In <$> inPorts0
+  outPorts1 = mkPort Out <$> outPorts0
+
+-- | Convert a 'LitHDL' to a 'TExpr'
+--
+-- __N.B.__: Clash 1.8 changed 'instDecl'\'s type signature. Where it would
+--           previously accept 'LitHDL' in its generics/parameters argument, it
+--           now accepts a 'TExpr'. This function is mostly there to ease this
+--           transition.
+litTExpr :: LitHDL -> TExpr
+litTExpr (B b) = TExpr Bool    (Literal Nothing (BoolLit b))
+litTExpr (S s) = TExpr String  (Literal Nothing (StringLit s))
+litTExpr (I i) = TExpr Integer (Literal Nothing (NumLit i))
+
+-- | Instantiate a component/entity in a block state
 instDecl
   :: forall backend
    . Backend backend
   => EntityOrComponent
   -- ^ Type of instantiation
   -> Identifier
-  -- ^ component/entity name
+  -- ^ Component/entity name
   -> Identifier
-  -- ^ instantiation label
-  -> [(Text, LitHDL)]
-  -- ^ attributes
+  -- ^ Instantiation label
   -> [(Text, TExpr)]
-  -- ^ in ports
+  -- ^ Generics / parameters
   -> [(Text, TExpr)]
-  -- ^ out ports
+  -- ^ In ports
+  -> [(Text, TExpr)]
+  -- ^ Out ports
   -> State (BlockState backend) ()
-instDecl entOrComp compName instLbl attrs inPorts outPorts = do
+instDecl entOrComp compName instLbl params inPorts outPorts = do
 
   inPorts' <- mapM (mkPort In) inPorts
   outPorts' <- mapM (mkPort Out) outPorts
 
   addDeclaration $
     InstDecl
-      entOrComp Nothing [] compName instLbl (mkAttrs attrs)
+      entOrComp Nothing [] compName instLbl (mkParams params)
       (NamedPortMap (inPorts' ++ outPorts'))
-    where
+   where
     mkPort
       :: PortDirection
       -> (Text, TExpr)
@@ -709,21 +915,12 @@ instDecl entOrComp compName instLbl attrs inPorts outPorts = do
       TExpr ty pExpr' <- toIdentifier (nmText <> "_port")  pExpr
       pure (Identifier (Id.unsafeMake nmText) Nothing, inOrOut, ty, pExpr')
 
-    -- Convert a list of name attributes to the form clash wants
-    mkAttrs :: [(Text.Text, LitHDL)] -> [(Expr, HWType, Expr)]
-    mkAttrs = map (\(s, ty) -> ( Identifier (Id.unsafeMake s) Nothing
-                               , hdlTy ty, litExpr ty) )
-
-    litExpr :: LitHDL -> Expr
-    litExpr (B b) = Literal Nothing (BoolLit b)
-    litExpr (S s) = Literal Nothing (StringLit s)
-    litExpr (I i) = Literal Nothing (NumLit i)
-
-    hdlTy :: LitHDL -> HWType
-    hdlTy = \case
-      B{} -> Bool
-      S{} -> String
-      I{} -> Integer
+    -- Convert a list of name generics / parameters to the form clash wants
+    mkParams :: [(Text.Text, TExpr)] -> [(Expr, HWType, Expr)]
+    mkParams = map $ \(paramName, texpr) ->
+      ( Identifier (Id.unsafeMake paramName) Nothing
+      , ety texpr
+      , eex texpr )
 
 -- | Wires the two given `TExpr`s together using a newly declared
 -- signal with (exactly) the given name `sigNm`. The new signal has an
@@ -736,14 +933,14 @@ viaAnnotatedSignal
   -- ^ expression the signal is assigned to
   -> TExpr
   -- ^ expression (must be identifier) to which the signal is assigned
-  -> [Attr']
+  -> [Attr Text]
   -- ^ the attributes to annotate the signal with
   -> State (BlockState backend) ()
 viaAnnotatedSignal sigNm (TExpr fromTy fromExpr) (TExpr toTy (Identifier outNm Nothing)) attrs
   | fromTy == toTy = do
       addDeclaration (NetDecl Nothing sigNm (Annotated attrs fromTy))
-      addDeclaration (Assignment sigNm fromExpr)
-      addDeclaration (Assignment outNm (Identifier sigNm Nothing))
+      addDeclaration (Assignment sigNm Cont fromExpr)
+      addDeclaration (Assignment outNm Cont (Identifier sigNm Nothing))
 viaAnnotatedSignal _ inTExpr outTExpr@(TExpr _ (Identifier _ _)) _ =
   error $ "viaAnnotatedSignal: The in and out expressions \"" <> show inTExpr <>
   "\" and \"" <> show outTExpr <> "\" have non-matching types."
@@ -771,7 +968,10 @@ toIdentifier'
   -- ^ identifier to expression
 toIdentifier' _ (TExpr _ (Identifier aExpr Nothing)) = pure aExpr
 toIdentifier' nm texp = do
-  ~(TExpr _ (Identifier nm' Nothing)) <- assign nm texp
+  t <- assign nm texp
+  let nm' = case t of
+              TExpr _ (Identifier x Nothing) -> x
+              t' -> error ("toIdentifier' expected an Identifier, but got: " <> show t')
   pure nm'
 
 -- | Get an identifier to an expression, creating a new assignment if
@@ -818,15 +1018,43 @@ andExpr nm a b = do
       SystemVerilog -> aIdent <> " && " <> bIdent
   assign nm $ TExpr Bool (Identifier (Id.unsafeMake andTxt) Nothing)
 
+-- | Massage a reset to work as active-high reset.
+unsafeToActiveHigh
+  :: Backend backend
+  => Text
+  -- ^ Name hint
+  -> TExpr
+  -- ^ Reset signal
+  -> State (BlockState backend) TExpr
+unsafeToActiveHigh nm rExpr = do
+  resetLevel <- vResetPolarity <$> liftToBlockState (getDomainConf (ety rExpr))
+  case resetLevel of
+    ActiveHigh -> pure rExpr
+    ActiveLow -> notExpr nm rExpr
+
+-- | Massage a reset to work as active-low reset.
+unsafeToActiveLow
+  :: Backend backend
+  => Text
+  -- ^ Name hint
+  -> TExpr
+  -- ^ Reset signal
+  -> State (BlockState backend) TExpr
+unsafeToActiveLow nm rExpr = do
+  resetLevel <- vResetPolarity <$> liftToBlockState (getDomainConf (ety rExpr))
+  case resetLevel of
+    ActiveLow -> pure rExpr
+    ActiveHigh -> notExpr nm rExpr
+
 -- | Negate @(not)@ an expression, assigning it to a new identifier.
 notExpr
   :: Backend backend
   => Text
   -- ^ name hint
   -> TExpr
-  -- ^ a
+  -- ^ @a@
   -> State (BlockState backend) TExpr
-  -- ^ not a
+  -- ^ @not a@
 notExpr _ T = pure F
 notExpr _ F = pure T
 notExpr nm aExpr = do
@@ -851,9 +1079,9 @@ pureToBV
   -> Int
   -- ^ Size (n)
   -> TExpr
-  -- ^ ARG
+  -- ^ @ARG@
   -> State (BlockState VHDLState) TExpr
-  -- ^ (0 to n => ARG)
+  -- ^ @(0 to n => ARG)@
 pureToBV nm n arg = do
   arg' <- Id.toText <$> toIdentifier' nm arg
   -- This is very hard coded and hacky
@@ -863,7 +1091,7 @@ pureToBV nm n arg = do
 -- | Creates a BV that produces the following vhdl:
 --
 -- @
---    std_logic_vector(resize(ARG, Size))
+--    std_logic_vector(resize(ARG, n))
 -- @
 --
 -- TODO: Implement for (System)Verilog
@@ -875,7 +1103,7 @@ pureToBVResized
   -> TExpr
   -- ^ ARG
   -> State (BlockState VHDLState) TExpr
-  -- ^ std_logic_vector(resize(ARG, Size))
+  -- ^ @std_logic_vector(resize(ARG, n))@
 pureToBVResized nm n arg = do
   arg' <- Id.toText <$> toIdentifier' nm arg
   -- This is very hard coded and hacky
