@@ -79,7 +79,10 @@ import           Clash.Core.Var
 import           Clash.Core.VarEnv
   (InScopeSet, extendInScopeSet, extendInScopeSetList, mkInScopeSet,
    uniqAway, uniqAway', mapVarEnv, eltsVarEnv, unitVarSet, emptyVarEnv,
-   mkVarEnv, eltsVarSet, elemVarEnv, lookupVarEnv, extendVarEnv, elemVarSet)
+   mkVarEnv, eltsVarSet, elemVarEnv, lookupVarEnv, extendVarEnv, elemVarSet,
+   differenceVarEnv)
+import           Clash.Data.UniqMap (UniqMap)
+import qualified Clash.Data.UniqMap as UniqMap
 import           Clash.Debug
 import           Clash.Driver.Types
   (TransformationInfo(..), DebugOpts(..), BindingMap, Binding(..), IsPrim(..),
@@ -554,9 +557,9 @@ liftBinding :: LetBinding
             -> RewriteMonad extra LetBinding
 liftBinding (var@Id {varName = idName} ,e) = do
   -- Get all local FVs, excluding the 'idName' from the let-binding
-  let unitFV :: Var a -> Const (UniqSet TyVar,UniqSet Id) (Var a)
-      unitFV v@(Id {})    = Const (emptyUniqSet,unitUniqSet (coerce v))
-      unitFV v@(TyVar {}) = Const (unitUniqSet (coerce v),emptyUniqSet)
+  let unitFV :: Var a -> Const (UniqMap TyVar,UniqMap Id) (Var a)
+      unitFV v@(Id {})    = Const (UniqMap.empty,UniqMap.singletonUnique (coerce v))
+      unitFV v@(TyVar {}) = Const (UniqMap.singletonUnique (coerce v),UniqMap.empty)
 
       interesting :: Var a -> Bool
       interesting Id {idScope = GlobalId} = False
@@ -565,8 +568,8 @@ liftBinding (var@Id {varName = idName} ,e) = do
 
       (boundFTVsSet,boundFVsSet) =
         getConst (Lens.foldMapOf (termFreeVars' interesting) unitFV e)
-      boundFTVs = eltsUniqSet boundFTVsSet
-      boundFVs  = eltsUniqSet boundFVsSet
+      boundFTVs = UniqMap.elems boundFTVsSet
+      boundFVs  = UniqMap.elems boundFVsSet
 
   -- Make a new global ID
   tcm       <- Lens.view tcCache
@@ -594,12 +597,12 @@ liftBinding (var@Id {varName = idName} ,e) = do
       newBody = mkTyLams (mkLams e' boundFVs) boundFTVs
 
   -- Check if an alpha-equivalent global binder already exists
-  aeqExisting <- (eltsUniqMap . filterUniqMap ((`aeqTerm` newBody) . bindingTerm)) <$> Lens.use bindings
+  aeqExisting <- (UniqMap.elems . UniqMap.filter ((`aeqTerm` newBody) . bindingTerm)) <$> Lens.use bindings
   case aeqExisting of
     -- If it doesn't, create a new binder
     [] -> do -- Add the created function to the list of global bindings
              let r = newBodyId `globalIdOccursIn` newBody
-             bindings %= extendUniqMap newBodyNm
+             bindings %= UniqMap.insert newBodyNm
                                     -- We mark this function as internal so that
                                     -- it can be inlined at the very end of
                                     -- the normalisation pipeline as part of the
@@ -608,7 +611,11 @@ liftBinding (var@Id {varName = idName} ,e) = do
                                     -- function at this moment for a reason!
                                     -- (termination, CSE and DEC oppertunities,
                                     -- ,etc.)
+#if MIN_VERSION_ghc(9,2,0)
+                                    (Binding newBodyId sp NoUserInlinePrag IsFun newBody r)
+#else
                                     (Binding newBodyId sp NoUserInline IsFun newBody r)
+#endif
              -- Return the new binder
              return (var, newExpr)
     -- If it does, use the existing binder
@@ -650,7 +657,7 @@ addGlobalBind
 addGlobalBind vNm ty sp inl body = do
   let vId = mkGlobalId ty vNm
       r = vId `globalIdOccursIn` body
-  (ty,body) `deepseq` bindings %= extendUniqMap vNm (Binding vId sp inl IsFun body r)
+  (ty,body) `deepseq` bindings %= UniqMap.insert vNm (Binding vId sp inl IsFun body r)
 
 -- | Create a new name out of the given name, but with another unique. Resulting
 -- unique is guaranteed to not be in the given InScopeSet.
@@ -672,7 +679,7 @@ cloneNameWithBindingMap
   -> m (Name a)
 cloneNameWithBindingMap binders nm = do
   i <- getUniqueM
-  return (uniqAway' (`elemUniqMapDirectly` binders) i (setUnique nm i))
+  return (uniqAway' (`UniqMap.elem` binders) i (setUnique nm i))
 
 {-# INLINE isUntranslatable #-}
 -- | Determine if a term cannot be represented in hardware
@@ -724,7 +731,7 @@ whnfRW
   -> Term
   -> Rewrite extra
   -> RewriteMonad extra Term
-whnfRW isSubj ctx@(TransformContext is0 _) e rw = do
+whnfRW isSubj ctx@(TransformContext is0 hist) e rw = do
   tcm <- Lens.view tcCache
   bndrs <- Lens.use bindings
   eval <- Lens.view evaluator
@@ -732,11 +739,18 @@ whnfRW isSubj ctx@(TransformContext is0 _) e rw = do
   let (ids1,ids2) = splitSupply ids
   uniqSupply Lens..= ids2
   gh <- Lens.use globalHeap
+  let lh = localBinders mempty hist
 
-  case whnf' eval bndrs tcm gh ids1 is0 isSubj e of
+  case whnf' eval bndrs lh tcm gh ids1 is0 isSubj e of
     (!gh1,ph,v) -> do
       globalHeap Lens..= gh1
-      bindPureHeap tcm ph rw ctx v
+      bindPureHeap tcm (ph `differenceVarEnv` lh) rw ctx v
+ where
+  localBinders acc [] = acc
+  localBinders !acc (h:hs) = case h of
+    LetBody ls -> localBinders (acc <> mkVarEnv ls) hs
+    _ -> localBinders acc hs
+
 {-# SCC whnfRW #-}
 
 -- | Binds variables on the PureHeap over the result of the rewrite
@@ -785,9 +799,9 @@ bindPureHeap tcm heap rw ctx0@(TransformContext is0 hist) e = do
   where
     heapIds = map fst bndrs
     is1 = extendInScopeSetList is0 heapIds
-    ctx = TransformContext is1 (LetBody heapIds : hist)
+    ctx = TransformContext is1 (LetBody bndrs : hist)
 
-    bndrs = map toLetBinding $ toListUniqMap heap
+    bndrs = map toLetBinding $ UniqMap.toList heap
 
     toLetBinding :: (Unique,Term) -> LetBinding
     toLetBinding (uniq,term) = (nm, term)
